Type TConfig
	Field GWidth
	Field GHeight
	Field FullScreen
	Field ViewRange
	Field AmbientOcclusion
	Field CursorEnabled
	Field MouseSensitivity#
	
	Field Name$
	Field IP$
	
	Field MapWidth
	Field MapHeight
	Field MapDepth
End Type

Const CONFIG_PATH$ = "Config.ini"

Function CreateDefaultConfig.TConfig()
	Local Config.TConfig = New TConfig
	
	Config\GWidth = 800
	Config\GHeight = 600
	Config\FullScreen = 0
	Config\ViewRange = 30
	Config\AmbientOcclusion = 1
	Config\CursorEnabled = 1
	Config\MouseSensitivity# = 5.0
	Config\Name$ = "Horst"
	Config\IP$ = "127.0.0.1"
	Config\MapWidth = 512
	Config\MapHeight = 64
	Config\MapDepth = 512
	
	Return Config
End Function

Function LoadConfig.TConfig()
	Local Instream = ReadFile( CONFIG_PATH$ )
	
	Local Config.TConfig = New TConfig
	
	If Instream Then
		While Not Eof( Instream )
			Local CurrLine$ = ReadLine( Instream )
			
			Local Offset = Instr( CurrLine$, "=" )
			
			If Offset Then
				Local LeftPart$ =  Trim( Left( CurrLine$, Offset - 1 ) )
				Local RightPart$ = Trim( Mid( CurrLine$, Offset + 1 ) )
				
				Select Lower( Trim( LeftPart$ ) )
					Case "graphicswidth"
						Config\GWidth = Int( RightPart$ )
					Case "graphicsheight"
						Config\GHeight = Int( RightPart$ )
					Case "fullscreen"
						Config\FullScreen = Int( RightPart$ )
					Case "visibilityrange"
						Config\ViewRange = Int( RightPart$ )
					Case "ambientocclusion"
						Config\AmbientOcclusion = Int( RightPart$ )
					Case "cursorenabled"
						Config\CursorEnabled = Int( RightPart$ )
					Case "playername"
						Config\Name$ = RightPart$
					Case "connectionip"
						Config\IP$ = RightPart$
					Case "mapwidth"
						Config\MapWidth = Int( RightPart$ )
					Case "mapheight"
						Config\MapHeight = Int( RightPart$ )
					Case "mapdepth"
						Config\MapDepth = Int( RightPart$ )
					Case "mousesensitivity"
						Config\MouseSensitivity# = Float( RightPart$ )
				End Select
			EndIf
		Wend
		
		CloseFile( Instream )
	Else
		Config = CreateDefaultConfig()
		
		SaveConfig( Config )
	EndIf
	
	Return Config
End Function

Function SaveConfig( Config.TConfig )
	Local Outstream = WriteFile( CONFIG_PATH$ )
	
	If Outstream Then
		WriteLine( Outstream, "GraphicsWidth = " + Config\GWidth )
		WriteLine( Outstream, "GraphicsHeight = " + Config\GHeight )
		WriteLine( Outstream, "Fullscreen = " + Config\FullScreen )
		WriteLine( Outstream, "VisibilityRange = " + Config\ViewRange )
		WriteLine( Outstream, "AmbientOcclusion = " + Config\AmbientOcclusion )
		WriteLine( Outstream, "CursorEnabled = " + Config\CursorEnabled )
		WriteLine( Outstream, "MouseSensitivity = " + Config\MouseSensitivity# )
		WriteLine( Outstream, "" )
		WriteLine( Outstream, "PlayerName = " + Config\Name$ )
		WriteLine( Outstream, "ConnectionIP = " + Config\IP$ )
		WriteLine( Outstream, "" )
		WriteLine( Outstream, "MapWidth = " + Config\MapWidth )
		WriteLine( Outstream, "MapHeight = " + Config\MapHeight )
		WriteLine( Outstream, "MapDepth = " + Config\MapDepth )
		
		CloseFile( Outstream )
	EndIf
End Function