Include "GUI\GUI_Constants.bb"

Global g_GUI_CurrentStyle.TGUIStyle, g_GUI_PickedIcon.TImageClass3D, g_GUI_PickedIconX, g_GUI_PickedIconY, g_GUI_PickedItembox.TItembox
Global g_GUI_PickedGadget.TGadget, g_GUI_ItemHover
Global ImageBank, LookupTable


Global g_GUI_Event_Window.TWindow
Global g_GUI_Event_Button.TGadget
Global g_GUI_Event_Checkbox.TGadget
Global g_GUI_Event_Tabber.TTabber
Global g_GUI_Event_RadioButton.TGadget
Global g_GUI_Event_Itembox.TItembox
Global g_GUI_Event_Combobox.TGadget
Global g_GUI_Event_Listbox.TGadget
Global g_GUI_Event_Textfield.TGadget


Dim Images.TImageClass3D( 500 )
Dim Fonts.TFont3D( 500 )

Type TGUIStyle
	Field Active.TGUIStyle
	Field Inactive.TGUIStyle
	Field Hover.TGUIStyle
	Field Pressed.TGUIStyle
	Field Tabber.TGUIStyle
	
	Field CheckBox.TImageClass3D
	Field RadioCircle.TImageClass3D
	Field Dropdown.TImageClass3D
	Field Highlight.TImageClass3D[ GUI_GADGET_COUNT ]
	Field Font.TFont3D[ GUI_GADGET_COUNT ]
	Field Fontsize[ GUI_GADGET_COUNT ]
	Field FontColor[ GUI_GADGET_COUNT ]
	
	Field Background.TImageClass3D[ GUI_GADGET_COUNT ]
	Field EdgeTL.TImageClass3D[ GUI_GADGET_COUNT ]
	Field EdgeTR.TImageClass3D[ GUI_GADGET_COUNT ]
	Field EdgeBL.TImageClass3D[ GUI_GADGET_COUNT ]
	Field EdgeBR.TImageClass3D[ GUI_GADGET_COUNT ]
	Field BorderT.TImageClass3D[ GUI_GADGET_COUNT ]
	Field BorderB.TImageClass3D[ GUI_GADGET_COUNT ]
	Field BorderR.TImageClass3D[ GUI_GADGET_COUNT ]
	Field BorderL.TImageClass3D[ GUI_GADGET_COUNT ]
End Type

Type TWindow
	Field X
	Field Y
	Field Width
	Field Height
	Field Caption$
	Field Alignment
	Field Flags
	
	Field TitlebarHeight
	
	Field Active
	Field Visible
	
	Field MouseOver
	Field MouseOverTitle
	
	Field Picked
	Field LastPickedX
	Field LastPickedY
	
	Field Alpha#
End Type

Type TGadget
	Field Kind
	Field X
	Field Y
	Field Width
	Field Height
	
	Field Caption$
	Field TextAlignment
	
	Field Visible
	Field Active
	
	Field MouseOver
	Field Pressed
	Field Clicked
	
	Field TabGroup
	Field Group
	
	Field GadgetData
	
	Field Hwnd.TWindow
End Type

Type TTabber
	Field Hwnd.TWindow
	Field X
	Field Y
	Field Width
	Field Height
	Field TabHeight
	
	Field GadgetCount
	Field Gadgets.TGadget[ 128 ]
	
	Field TabCount
	Field Tabs.TGadget[ 128 ]
End Type

Type TItemBox
	Field Hwnd.TWindow
	Field X
	Field Y
	Field Width
	Field Height
	
	Field IconBank
	Field GadgetBank
	
	Field BoxSize
	Field Padding
End Type

Function SetGuiStyle( Style.TGUIStyle )
	g_GUI_CurrentStyle = Style
	
	For Gadget.TGadget = Each TGadget
		If Gadget\Kind = GUI_T_LABEL Then RedistributeLabel( Gadget, Style )
	Next
End Function

Function CreateWindow.TWindow( X, Y, Width, Height, Caption$ = "", TextAlignment = GUI_A_LEFT, Flags = GUI_W_TITLEBAR )
	Window.TWindow = New TWindow
		Window\X = X
		Window\Y = Y
		Window\Width = Width
		Window\Height = Height
		Window\Caption$ = Caption$
		Window\Alignment = TextAlignment
		Window\Visible = True
		Window\Active = True
		Window\Flags = Flags
		Window\Alpha# = 1
	
	MoveWindowToFront( Window )
	
	Return Window
End Function

Function CreateLabel.TGadget( Hwnd.TWindow, X, Y, Width, Height, Caption$ = "", TextAlignment = GUI_A_LEFT )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_LABEL
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\Caption$ = Caption$
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
	
	RedistributeLabel( Gadget, g_GUI_CurrentStyle )
	
	Return Gadget
End Function

Function CreateButton.TGadget( Hwnd.TWindow, X, Y, Width, Height, Caption$ = "", TextAlignment = GUI_A_CENTER )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_BUTTON
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\Caption$ = Caption$
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
	
	Return Gadget
End Function

Function CreateCheckbox.TGadget( Hwnd.TWindow, X, Y, Width, Height, Caption$ = "", TextAlignment = GUI_A_LEFT )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_CHECKBOX
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\Caption$ = Caption$
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
	
	Return Gadget
End Function

Function CreateTabber.TTabber( Hwnd.TWindow, X, Y, Width, Height, TabHeight = 20 )
	Tabber.TTabber = New TTabber
		Tabber\Hwnd = Hwnd
		Tabber\X = X
		Tabber\Y = Y
		Tabber\Width = Width
		Tabber\Height = Height
		Tabber\TabHeight = TabHeight
	
	Return Tabber
End Function

Function CreateRadioButton.TGadget( Hwnd.TWindow, X, Y, Width, Height, Caption$ = "", Group = 0, TextAlignment = GUI_A_LEFT )
	Gadget.Tgadget = New TGadget
		Gadget\Kind = GUI_T_RADIOBUTTON
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\Caption$ = Caption$
		Gadget\Group = Group
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
	
	ActivateRadioButton( Gadget )
	
	Return Gadget
End Function

Function CreateItembox.TItembox( Hwnd.TWindow, X, Y, Width, Height, BoxSize = 50, Padding = 4, TextAlignment = GUI_A_RIGHT )
	Itembox.TItembox = New TItembox
		Itembox\Hwnd = Hwnd
		Itembox\X = X
		Itembox\Y = Y
		Itembox\Width = Width
		Itembox\Height = Height
		Itembox\BoxSize = BoxSize
		Itembox\Padding = Padding
		Itembox\Iconbank = CreateBank( Width*Height*4 )
		Itembox\Gadgetbank = CreateBank( Width*Height*4 )
	
	For i = 0 To Width - 1
		For t = 0 To Height - 1
			Gadget.TGadget = New TGadget
				Gadget\Kind = GUI_T_ITEMBOX
				Gadget\Hwnd = Hwnd
				Gadget\X = X + i*( BoxSize + Padding )
				Gadget\Y = Y + t*( BoxSize + Padding )
				Gadget\Width = BoxSize
				Gadget\Height = BoxSize
				Gadget\TextAlignment = TextAlignment
				Gadget\Group = Handle( ItemBox )
				Gadget\Visible = True
				Gadget\GadgetData = CreateBank( 2 )
			
			PokeInt Itembox\Gadgetbank, ( i + t*Width )*4, Handle( Gadget )
		Next
	Next
	
	Return Itembox
End Function

Function CreateComboBox.TGadget( Hwnd.TWindow, X, Y, Width, Height )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_COMBOBOX
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\GadgetData = CreateBank( 6 )
		Gadget\Visible = True
	
	PokeShort Gadget\GadgetData, 2, Height
	
	Return Gadget
End Function

Function CreateListbox.TGadget( Hwnd.TWindow, X, Y, Width, EntryHeight, EntryCount )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_LISTBOX
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = EntryHeight*EntryCount
		Gadget\GadgetData = CreateBank( 6 )
		Gadget\Visible = True
	
	PokeShort Gadget\GadgetData, 4, EntryHeight
	
	Return Gadget
End Function

Function CreateTextfield.TGadget( Hwnd.TWindow, X, Y, Width, Height, Caption$ = "", TextAlignment = GUI_A_LEFT )
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_TEXTFIELD
		Gadget\Hwnd = Hwnd
		Gadget\X = X
		Gadget\Y = Y
		Gadget\Width = Width
		Gadget\Height = Height
		Gadget\Caption$ = Caption$
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
		Gadget\GadgetData = CreateBank( 4 )
	
	Return Gadget
End Function

Function DestroyWindow( Window.TWindow )
	For Gadget.TGadget = Each TGadget
		If Gadget\Hwnd = Window Then Delete Gadget
	Next
	
	Delete Window
End Function

Function RenderGUI()
	Window.TWindow = Last TWindow
	
	While Window <> Null
		If Window\Visible Then
			Origin3D 0, 0
			
			RenderWindow( Window, g_GUI_CurrentStyle )
			
			If Window\Active Then CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Active Else CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Inactive
			Origin3D Window\X, Window\Y + Imageheight3D( CurrStyle\BorderT[ GUI_T_WINDOW ] )
			
			For Tabber.TTabber = Each TTabber
				If Tabber\Hwnd = Window Then
					For i = 0 To Tabber\TabCount - 1
						RenderTab( Tabber\Tabs[ i ], g_GUI_CurrentStyle )
					Next
					
					DrawGadgetRect g_GUI_CurrentStyle\Tabber, GUI_T_TAB, Tabber\X, Tabber\Y + Tabber\TabHeight, Tabber\Width, Tabber\Height - Tabber\TabHeight
				EndIf
			Next
			
			For Gadget.TGadget = Each TGadget
				If Gadget\Hwnd = Window And Gadget\Visible Then
					Select Gadget\Kind
						Case GUI_T_LABEL
							RenderLabel( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_BUTTON
							RenderButton( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_CHECKBOX
							RenderCheckbox( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_RADIOBUTTON
							RenderRadioButton( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_ITEMBOX
							RenderItemBox( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_LISTBOX
							RenderListBox( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_COMBOBOX
							RenderComboBox( Gadget, g_GUI_CurrentStyle )
						Case GUI_T_TEXTFIELD
							RenderTextfield( Gadget, g_GUI_CurrentStyle )
					End Select
				EndIf
			Next
		EndIf
		
		Window = Before Window
	Wend
	
	Origin3D 0, 0
	Color3D 255, 255, 255
	
	If g_GUI_PickedIcon <> Null Then RectImage3D g_GUI_PickedIcon, MouseX(), MouseY(), g_GUI_PickedItembox\BoxSize, g_GUI_PickedItembox\BoxSize
End Function

Function UpdateGUI()
	;------- Einige Werte zurcksetzen
	
	For Window.TWindow = Each TWindow
		Window\MouseOver = False
		Window\MouseOverTitle = False
		If Not MD1 Then Window\Picked = False
	Next
	
	For Gadget.TGadget = Each TGadget
		Gadget\MouseOver = False
		If Gadget\Kind <> GUI_T_CHECKBOX And Gadget\Kind <> GUI_T_TAB And Gadget\Kind <> GUI_T_RADIOBUTTON And Gadget\Kind <> GUI_T_COMBOBOX Then Gadget\Pressed = False
		Gadget\Clicked = False
	Next
	
	g_GUI_Event_Window.TWindow		= Null
	g_GUI_Event_Button.TGadget		= Null
	g_GUI_Event_Checkbox.TGadget	= Null
	g_GUI_Event_Tabber.TTabber		= Null
	g_GUI_Event_RadioButton.TGadget	= Null
	g_GUI_Event_Itembox.TItembox	= Null
	g_GUI_Event_Combobox.TGadget	= Null
	g_GUI_Event_Listbox.TGadget		= Null
	
	g_GUI_ItemHover = 0
	
	;------- Zuerst prfen, ob die Maus drberschwebt
	
	For Window.TWindow = Each TWindow
		If Window\Active Then CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Active Else CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Inactive
		
		If Window\Visible Then
			Window\MouseOver = RectsOverlap( Window\X, Window\Y, Window\Width, Window\Height, MX, MY, 1, 1 )
			If ( Window\Flags And GUI_W_TITLEBAR ) Then Window\MouseOverTitle = RectsOverlap( Window\X, Window\Y, Window\Width, ImageHeight3D( CurrStyle\BorderT[ GUI_T_WINDOW ] ), MX, MY, 1, 1 )
			
			If Window\MouseOver Then Exit
		EndIf
	Next
	
	Gadget.TGadget = Last TGadget
	While Gadget <> Null
		If Gadget\Hwnd\Visible And Gadget\Hwnd\Active Then
			If Gadget\Visible Then
				If Gadget\Hwnd\Active Then CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Active Else CurrStyle.TGUIStyle = g_GUI_CurrentStyle\Inactive
				Gadget\MouseOver = RectsOverlap( Gadget\Hwnd\X + Gadget\X, Gadget\Hwnd\Y + Gadget\Y + ImageHeight3D( CurrStyle\BorderT[ GUI_T_WINDOW ] ), Gadget\Width, Gadget\Height, MX, MY, 1, 1 )
				
				If Gadget\Kind = GUI_T_ITEMBOX And Gadget\MouseOver Then g_GUI_ItemHover = Gadget\GadgetData
				
				If Gadget\MouseOver Then Exit
			EndIf
		EndIf
		
		Gadget = Before Gadget
	Wend
	
	;------- Dann Mausklicks abfragen
	
	If MH1 Then
		For Window.TWindow = Each TWindow
			If Window\MouseOver Then
				Window\Active = True
				Insert Window Before First TWindow
				
				If Window\MouseOverTitle And ( Not( Window\Flags And GUI_W_LOCKED ) ) Then
					Window\Picked = True
					Window\LastPickedX = MX
					Window\LastPickedY = MY
				EndIf
				
				g_GUI_Event_Window = Window
			Else
				Window\Active = False
			EndIf
		Next
		
		For Gadget.TGadget = Each TGadget
			If Gadget\MouseOver Then
				Gadget\Clicked = True
				
				TriggerGadget( Gadget )
			ElseIf Gadget\Kind = GUI_T_COMBOBOX Then
				ComboboxDropup( Gadget )
			ElseIf Gadget\Kind = GUI_T_TEXTFIELD Then
				Gadget\Active = False
			EndIf
		Next
	EndIf
	
	;------- Einige Gadgets reagieren noch auf die gedrckte Maustaste, die werden hier gleich auch noch bedient 
	
	If MD1 Then
		For Window.TWindow = Each TWindow
			If Window\Picked Then
				Window\X = Window\X + ( MX - Window\LastPickedX )
				Window\Y = Window\Y + ( MY - Window\LastPickedY )
				Window\LastPickedX = MX
				Window\LastPickedY = MY
			EndIf
		Next
		
		For Gadget.TGadget = Each TGadget
			If Gadget\MouseOver And Gadget\Kind <> GUI_T_CHECKBOX And Gadget\Kind <> GUI_T_TAB And Gadget\Kind <> GUI_T_RADIOBUTTON And Gadget\Kind <> GUI_T_COMBOBOX Then
				Gadget\Pressed = True
			EndIf
		Next
	Else
		If g_GUI_PickedIcon <> Null Then DropItem()
	EndIf
	
	;------- Tasten abfangen - betrifft nur das Textfeld
	
	For Gadget.TGadget = Each TGadget
		If Gadget\Kind = GUI_T_TEXTFIELD And Gadget\Active Then
			Local Char = Asc( eg_getkey() )
			
			If Char > 31 Then
				Gadget\Caption$ = Gadget\Caption$ + Chr( Char )
				
				PokeInt Gadget\GadgetData, 0, MilliSecs()
			ElseIf Char = 17 Then
				Local NewLength = Len( Gadget\Caption$ ) - 1
				
				If NewLength >= 0 Then Gadget\Caption$ = Left( Gadget\Caption$, NewLength )
				
				PokeInt Gadget\GadgetData, 0, MilliSecs()
			ElseIf Char = 19 Then
				Gadget\Caption$ = Gadget\Caption$ + BTGetClipboardText()
			EndIf
			
			Exit
		EndIf
	Next
End Function

;Krass aus dieser einen GUI geklaut.

Global eg_keyb_repeattime	= 20		;--> Keyboard Repetierzeit (in ms)
Global eg_keyb_timetorepeat	= 300		;--> Zeit bis zum Einsetzen des Repeatiervorgangs (in ms)
Global eg_keyb_timer						;--> Zeitmesser fr Keyboard

Function eg_getkey$()
;==============================================================================================================
; ...liefert CHR$(0) wenn keine Taste gedrckt wurde. Ansonsten wird das ASCII Zeichen zurckgeliefert.
;
; Die Routine ersetzt im Prinzip die GetKey() Funktion. Allerdings ermittelt diese Funktion den Tastenstatus.
; Somit kann eine repetierende Abfrage erzeugt werden. Neben den 'normalen' Tasten wie A,B,C usw. werden auch
; einige Steuertasten (links, rechts usw.) abgefragt. Ebenso werden die Kombinationen mit Shift, Strg oder Alt
; ausgewertet und ein entsprechendes ASCII Zeichen zurckgegeben.
;
; Die Steuertasten geben folgende ASCII Werte zurck:
; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
; 01 - Pos1
; 02 - Ende
; 03 - Shift + Pos1
; 04 - Shift + Ende
; 05 - Links
; 06 - Rechts
; 07 - Shift + Links
; 08 - Shift + Rechts
; 09 - Strg + Links
; 10 - Strg + Rechts
; 11 - Strg + Shift + Links
; 12 - Strg + Shift + Rechts
; 13 - Return
; 14 - Entfernen
; 15 - TAB
; 16 - Shift + TAB
; 17 - BackSpace (DEL)
; 18 - Strg + C und Strg + Einfg (kopieren)
; 19 - Strg + V und Shift + Einfg (einfgen)
; 20 - Strg + X (ausschneiden)
; 21 - Einfg

; 23 - hoch
; 24 - runter
; 25 - Bild hoch
; 26 - Bild runter
; 31 - ESC
;==============================================================================================================

	If KeyDown(42) Or KeyDown(54) Then eg_shift=1 Else eg_shift=0
	If KeyDown(29) Or KeyDown(157) Then eg_ctrl=1 Else eg_ctrl=0
	If KeyDown(56) Or KeyDown(184) Then eg_alt=1 Else eg_alt=0

	;--> erfasse den ersten Tastendruck ber GetKey() bzw. ber KeyHit() beim Ziffernblock
	eg_key$ = Chr$(GetKey())
	;--> konvertiere Ziffernblock
	If KeyHit(82)  Then eg_key$="0"
	If KeyHit(79)  Then eg_key$="1"
	If KeyHit(80)  Then eg_key$="2"
	If KeyHit(81)  Then eg_key$="3"
	If KeyHit(75)  Then eg_key$="4"
	If KeyHit(76)  Then eg_key$="5"
	If KeyHit(77)  Then eg_key$="6"
	If KeyHit(71)  Then eg_key$="7"
	If KeyHit(72)  Then eg_key$="8"
	If KeyHit(73)  Then eg_key$="9"
	If KeyHit(181) Then eg_key$="/"
	If KeyHit(55)  Then eg_key$="*"
	If KeyHit(74)  Then eg_key$="-"
	If KeyHit(78)  Then eg_key$="+"
	If KeyHit(83)  Then eg_key$="."
	
	;--> Steuertasten
	If eg_shift Then
		If KeyHit(203) Then eg_key$=Chr$(07)    ;links
		If KeyHit(205) Then eg_key$=Chr$(08)    ;rechts
		If KeyHit(199) Then eg_key$=Chr$(03)    ;Pos1
		If KeyHit(207) Then eg_key$=Chr$(04)    ;Ende
		If KeyHit(015) Then eg_key$=Chr$(16)    ;TAB
		If KeyHit(210) Then eg_key$=Chr$(19)    ;Einfg
		If eg_ctrl Then
			If eg_key$=Chr$(07) Then eg_key$=Chr$(11)	;links
			If eg_key$=Chr$(08) Then eg_key$=Chr$(12)	;rechts
		End If
	Else
		If KeyHit(203) Then eg_key$=Chr$(05)    ;links
		If KeyHit(205) Then eg_key$=Chr$(06)    ;rechts
		If KeyHit(200) Then eg_key$=Chr$(23)    ;hoch
		If KeyHit(208) Then eg_key$=Chr$(24)    ;runter
		If KeyHit(201) Then eg_key$=Chr$(25)    ;Bild hoch
		If KeyHit(209) Then eg_key$=Chr$(26)    ;Bild runter
		If KeyHit(199) Then eg_key$=Chr$(01)    ;Pos1
		If KeyHit(207) Then eg_key$=Chr$(02)    ;Ende
		If KeyHit(015) Then eg_key$=Chr$(15)    ;TAB
		If KeyHit(210) Then eg_key$=Chr$(21)    ;Einfg
		If eg_ctrl Then
			If eg_key$=Chr$(05) Then eg_key$=Chr$(09)	;links
			If eg_key$=Chr$(06) Then eg_key$=Chr$(10)	;rechts
			If KeyHit(46) Then eg_key$=Chr$(18)      ;Strg+C
			If KeyHit(47) Then eg_key$=Chr$(19)      ;Strg+V
			If KeyHit(45) Then eg_key$=Chr$(20)      ;Strg+X
			If eg_key$=Chr$(21) Then eg_key$=Chr$(18);Strg+Einfg
		End If
		;If KeyHit(001) Then eg_key$=Chr$(31)    ;ESC
		If KeyHit(211) Then eg_key$=Chr$(14)    ;Entf
		If KeyHit(14)  Then eg_key$=Chr$(17)    ;Del
		;If KeyHit(028) Then eg_key$=Chr$(13)    ;Return
		;If KeyHit(156) Then eg_key$=Chr$(13)    ;Return
	End If
	
	;--> wenn GetKey bzw. KeyHit() nichts ergeben haben, dann hlt der User womglich eine Taste fest
	;--> ermittle die Taste und frage sie ab. Der Timer dient der Repetiergeschwindigkeit.
	If eg_key$=Chr$(0) Then
		If eg_keyb_timer<MilliSecs() Then
			eg_keyb_timer=MilliSecs() + eg_keyb_repeattime
			
			;--> Steuertasten
			If eg_shift Then
				If KeyDown(203) Then eg_key$=Chr$(07)    ;links
				If KeyDown(205) Then eg_key$=Chr$(08)    ;rechts
				If eg_ctrl Then
					If KeyDown(203) Then eg_key$=Chr$(11)	;links
					If KeyDown(205) Then eg_key$=Chr$(12)	;rechts
				End If
			Else
				If KeyDown(203) Then eg_key$=Chr$(05)    ;links
				If KeyDown(205) Then eg_key$=Chr$(06)    ;rechts
				If KeyDown(200) Then eg_key$=Chr$(23)    ;hoch
				If KeyDown(208) Then eg_key$=Chr$(24)    ;runter
				If KeyDown(201) Then eg_key$=Chr$(25)    ;Bild hoch
				If KeyDown(209) Then eg_key$=Chr$(26)    ;Bild runter
				If eg_ctrl Then
					If KeyDown(203) Then eg_key$=Chr$(09)	;links
					If KeyDown(205) Then eg_key$=Chr$(10)	;rechts
				End If
				If KeyDown(211) Then eg_key$=Chr$(14)    ;Entf
				If KeyDown(14)  Then eg_key$=Chr$(17)    ;Del
				If KeyDown(57)  Then eg_key$=Chr$(32)    ;Leertaste
			End If
			
			;--> 0...9
			If eg_shift Then
				If KeyDown(11)  Then eg_key$="="      ;=
				If KeyDown(2)   Then eg_key$="!"      ;!
				If KeyDown(3)   Then eg_key$=Chr$(34) ;"
				If KeyDown(4)   Then eg_key$=""      ;
				If KeyDown(5)   Then eg_key$="$"      ;$
				If KeyDown(6)   Then eg_key$="%"      ;%
				If KeyDown(7)   Then eg_key$="&"      ;&
				If KeyDown(8)   Then eg_key$="/"      ;/
				If KeyDown(9)   Then eg_key$="("      ;(
				If KeyDown(10)  Then eg_key$=")"      ;)
			Else
				If KeyDown(11)  Then eg_key$="0"      ;0
				If KeyDown(2)   Then eg_key$="1"      ;1
				If KeyDown(3)   Then eg_key$="2"      ;2
				If KeyDown(4)   Then eg_key$="3"      ;3
				If KeyDown(5)   Then eg_key$="4"      ;4
				If KeyDown(6)   Then eg_key$="5"      ;5
				If KeyDown(7)   Then eg_key$="6"      ;6
				If KeyDown(8)   Then eg_key$="7"      ;7
				If KeyDown(9)   Then eg_key$="8"      ;8
				If KeyDown(10)  Then eg_key$="9"      ;9
			End If
			
			;--> Ziffernblock
			If KeyDown(82)  Then eg_key$="0"         ;0
			If KeyDown(79)  Then eg_key$="1"         ;1
			If KeyDown(80)  Then eg_key$="2"         ;2
			If KeyDown(81)  Then eg_key$="3"         ;3
			If KeyDown(75)  Then eg_key$="4"         ;4
			If KeyDown(76)  Then eg_key$="5"         ;5
			If KeyDown(77)  Then eg_key$="6"         ;6
			If KeyDown(71)  Then eg_key$="7"         ;7
			If KeyDown(72)  Then eg_key$="8"         ;8
			If KeyDown(73)  Then eg_key$="9"         ;9
			If KeyDown(181) Then eg_key$="/"         ;/
			If KeyDown(55)  Then eg_key$="*"         ;*
			If KeyDown(74)  Then eg_key$="-"         ;-
			If KeyDown(78)  Then eg_key$="+"         ;+
			If KeyDown(83)  Then eg_key$=","         ;,
			
			;--> a...z,
			If eg_shift Then
				If KeyDown(30)  Then eg_key$="A"      ;A
				If KeyDown(48)  Then eg_key$="B"      ;B
				If KeyDown(46)  Then eg_key$="C"      ;C
				If KeyDown(32)  Then eg_key$="D"      ;D
				If KeyDown(18)  Then eg_key$="E"      ;E
				If KeyDown(33)  Then eg_key$="F"      ;F
				If KeyDown(34)  Then eg_key$="G"      ;G
				If KeyDown(35)  Then eg_key$="H"      ;H
				If KeyDown(23)  Then eg_key$="I"      ;I
				If KeyDown(36)  Then eg_key$="J"      ;J
				If KeyDown(37)  Then eg_key$="K"      ;K
				If KeyDown(38)  Then eg_key$="L"      ;L
				If KeyDown(50)  Then eg_key$="M"      ;M
				If KeyDown(49)  Then eg_key$="N"      ;N
				If KeyDown(24)  Then eg_key$="O"      ;O
				If KeyDown(25)  Then eg_key$="P"      ;P
				If KeyDown(16)  Then eg_key$="Q"      ;Q
				If KeyDown(19)  Then eg_key$="R"      ;R
				If KeyDown(31)  Then eg_key$="S"      ;S
				If KeyDown(20)  Then eg_key$="T"      ;T
				If KeyDown(22)  Then eg_key$="U"      ;U
				If KeyDown(47)  Then eg_key$="V"      ;V
				If KeyDown(17)  Then eg_key$="W"      ;W
				If KeyDown(45)  Then eg_key$="X"      ;X
				If KeyDown(44)  Then eg_key$="Y"      ;Y
				If KeyDown(21)  Then eg_key$="Z"      ;Z
				If KeyDown(40)  Then eg_key$=""      ;
				If KeyDown(39)  Then eg_key$=""      ;
				If KeyDown(26)  Then eg_key$=""      ;
				If KeyDown(12)  Then eg_key$="?"      ;?
			Else
				If KeyDown(30)  Then eg_key$="a"      ;a
				If KeyDown(48)  Then eg_key$="b"      ;b
				If KeyDown(46)  Then eg_key$="c"      ;c
				If KeyDown(32)  Then eg_key$="d"      ;d
				If KeyDown(18)  Then eg_key$="e"      ;e
				If KeyDown(33)  Then eg_key$="f"      ;f
				If KeyDown(34)  Then eg_key$="g"      ;g
				If KeyDown(35)  Then eg_key$="h"      ;h
				If KeyDown(23)  Then eg_key$="i"      ;i
				If KeyDown(36)  Then eg_key$="j"      ;j
				If KeyDown(37)  Then eg_key$="k"      ;k
				If KeyDown(38)  Then eg_key$="l"      ;l
				If KeyDown(50)  Then eg_key$="m"      ;m
				If KeyDown(49)  Then eg_key$="n"      ;n
				If KeyDown(24)  Then eg_key$="o"      ;o
				If KeyDown(25)  Then eg_key$="p"      ;p
				If KeyDown(16)  Then eg_key$="q"      ;q
				If KeyDown(19)  Then eg_key$="r"      ;r
				If KeyDown(31)  Then eg_key$="s"      ;s
				If KeyDown(20)  Then eg_key$="t"      ;t
				If KeyDown(22)  Then eg_key$="u"      ;u
				If KeyDown(47)  Then eg_key$="v"      ;v
				If KeyDown(17)  Then eg_key$="w"      ;w
				If KeyDown(45)  Then eg_key$="x"      ;x
				If KeyDown(44)  Then eg_key$="y"      ;y
				If KeyDown(21)  Then eg_key$="z"      ;z
				If KeyDown(40)  Then eg_key$=""      ;
				If KeyDown(39)  Then eg_key$=""      ;
				If KeyDown(26)  Then eg_key$=""      ;
				If KeyDown(12)  Then eg_key$=""      ;
			End If
			
			;--> Sonderzeichen
			If eg_shift Then
				If KeyDown(13)  Then eg_key$="`"      ;`
				If KeyDown(86)  Then eg_key$=">"      ;>
				If KeyDown(51)  Then eg_key$=";"      ;;
				If KeyDown(52)  Then eg_key$=":"      ;:
				If KeyDown(53)  Then eg_key$="_"      ;_
				If KeyDown(43)  Then eg_key$="'"      ;'
				If KeyDown(41)  Then eg_key$=""      ;
				If KeyDown(27)  Then eg_key$="*"      ;*
			Else
				If KeyDown(13)  Then eg_key$=""      ;
				If KeyDown(86)  Then eg_key$="<"      ;<
				If KeyDown(51)  Then eg_key$=","      ;,
				If KeyDown(52)  Then eg_key$="."      ;.
				If KeyDown(53)  Then eg_key$="-"      ;-
				If KeyDown(43)  Then eg_key$="#"      ;#
				If KeyDown(41)  Then eg_key$="^"      ;^
				If KeyDown(27)  Then eg_key$="+"      ;+
			End If
			
			;--> Tasten die mit ALT-erreicht werden
			If eg_alt Then
				If KeyDown(11)  Then eg_key$="}"      ;}
				If KeyDown(3)   Then eg_key$=""      ;
				If KeyDown(4)   Then eg_key$=""      ;
				If KeyDown(8)   Then eg_key$="{"      ;{
				If KeyDown(9)   Then eg_key$="["      ;[
				If KeyDown(10)  Then eg_key$="]"      ;]
				If KeyDown(12)  Then eg_key$="\"      ;\
				If KeyDown(16)  Then eg_key$="@"      ;@
				If KeyDown(27)  Then eg_key$="~"      ;~
				If KeyDown(86)  Then eg_key$="|"      ;|
				If KeyDown(18)  Then eg_key$=""      ;Euro
				If KeyDown(50)  Then eg_key$=""      ;
			End If
			
		End If
	Else
		;--> Verzgerungszeit bis zum Einsetzen des Tastaur-Repeat
		eg_keyb_timer=MilliSecs()+eg_keyb_timetorepeat
	End If
	
 	Return eg_key$
End Function

Function TriggerGadget( Gadget.TGadget )
	Select Gadget\Kind
		Case GUI_T_CHECKBOX
			Gadget\Pressed = Not Gadget\Pressed
			g_GUI_Event_Checkbox = Gadget
		Case GUI_T_TAB
			ActivateTab( Gadget )
			g_GUI_Event_Tabber = Object.TTabber( Gadget\Group )
		Case GUI_T_RADIOBUTTON
			ActivateRadioButton( Gadget )
			g_GUI_Event_RadioButton = Gadget
		Case GUI_T_ITEMBOX
			PickItem( Gadget )
			g_GUI_Event_Itembox = Object.TItembox( Gadget\Group )
		Case GUI_T_COMBOBOX
			ComboboxDropdown( Gadget )
			If Gadget\Pressed Then g_GUI_Event_Combobox = Gadget
		Case GUI_T_LISTBOX
			ListboxPickEntry( Gadget )
			g_GUI_Event_Listbox = Gadget
		Case GUI_T_BUTTON
			g_GUI_Event_Button = Gadget
		Case GUI_T_TEXTFIELD
			Gadget\Active = True
			PokeInt Gadget\GadgetData, 0, MilliSecs()
			g_GUI_Event_Textfield = Gadget
	End Select
End Function

Function RenderWindow( Window.TWindow, Style.TGUIStyle )
	If Window\Active Then Style = Style\Active Else Style = Style\Inactive
	
	Window\TitlebarHeight = ImageHeight3D( Style\BorderT[ GUI_T_WINDOW ] )
	
	DrawGadgetRect( Style, GUI_T_WINDOW, Window\X, Window\Y, Window\Width, Window\Height )
	
	SetRGB Style\FontColor[ GUI_T_WINDOW ]
	SetFont3D Style\Font[ GUI_T_WINDOW ]
	
	If ( Window\Flags And GUI_W_TITLEBAR ) Then
		SetRGB Style\FontColor[ GUI_T_WINDOW ]
		SetFont3D Style\Font[ GUI_T_WINDOW ]
		
		Select Window\Alignment
			Case GUI_A_LEFT
				Text3D Window\X + ImageWidth3D( Style\EdgeTL[ GUI_T_WINDOW ] ), Window\Y + ImageHeight3D( Style\BorderT[ GUI_T_WINDOW ] )/2, Window\Caption$, False, True
			Case GUI_A_CENTER
				Text3D Window\X + Window\Width/2, Window\Y + ImageHeight3D( Style\BorderT[ GUI_T_WINDOW ] )/2, Window\Caption$, True, True
			Case GUI_A_RIGHT
				Text3D Window\X + Window\Width - ImageWidth3D( Style\EdgeTL[ GUI_T_WINDOW ] ) - StringWidth3D( Window\Caption$ ), Window\Y + ImageHeight3D( Style\BorderT[ GUI_T_WINDOW ] )/2, Window\Caption$, False, True
		End Select
	EndIf
End Function

Function RenderLabel( Gadget.TGadget, Style.TGUIStyle )
	Local TextLine$[ 512 ], Counter = 0, Caption$ = Gadget\Caption$
	
	Style = Style\Active
	
	While 1
		TextLine[ Counter ] = Caption$
		
		Offset = Instr( Caption$, Chr( 10 ) )
		
		If Offset Then
			TextLine[ Counter ] = Left( Caption$, Offset - 1 )
			TextLine[ Counter + 1 ] = Right( Caption$, Len( Caption$ ) - Offset )
		Else
			Exit
		EndIf
		
		Caption$ = TextLine[ Counter + 1 ]
		
		Counter = Counter + 1
	Wend
	
	SetRGB Style\FontColor[ GUI_T_LABEL ]
	SetFont3D Style\Font[ GUI_T_LABEL ]
	
	For i = 0 To Counter
		Select Gadget\TextAlignment
			Case GUI_A_LEFT
				Text3D Gadget\X, Gadget\Y + Style\Fontsize[ GUI_T_LABEL ]*i, TextLine[ i ]
			Case GUI_A_CENTER
				Text3D Gadget\X + Gadget\Width/2, Gadget\Y + Style\Fontsize[ GUI_T_LABEL ]*i, TextLine[ i ], True
			Case GUI_A_RIGHT
				Text3D Gadget\X + Gadget\Width - StringWidth3D( TextLine[ i ] ), Gadget\Y + Style\Fontsize[ GUI_T_LABEL ]*i, TextLine[ i ]
		End Select
	Next
End Function

Function RenderButton( Gadget.TGadget, Style.TGUIStyle )
	If Gadget\Pressed Then Style = Style\Pressed ElseIf Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	DrawGadgetRect Style, GUI_T_BUTTON, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	SetRGB Style\FontColor[ GUI_T_BUTTON ]
	SetFont3D Style\Font[ GUI_T_BUTTON ]
	
	If Gadget\Pressed Then
		OffsetX = 1
		OffsetY = 1
	EndIf
	
	Select Gadget\TextAlignment
		Case GUI_A_LEFT
			Text3D Gadget\X + 5 + OffsetX, Gadget\Y + Gadget\Height/2 + OffsetY, Gadget\Caption$, False, True
		Case GUI_A_CENTER
			Text3D Gadget\X + OffsetX + Gadget\Width/2, Gadget\Y + OffsetY + Gadget\Height/2, Gadget\Caption$, True, True
		Case GUI_A_RIGHT
			Text3D Gadget\X + Gadget\Width - StringWidth3D( Gadget\Caption$ ) - 5 + OffsetX, Gadget\Y + Gadget\Height/2 + OffsetY, Gadget\Caption$, False, True
	End Select
End Function

Function RenderCheckbox( Gadget.TGadget, Style.TGUIStyle )
	Pressed.TGUIStyle = Style\Pressed
	If Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	Color3D 255, 255, 255
	
	DrawImage3D Style\CheckBox, Gadget\X, Gadget\Y
	
	If Gadget\Pressed Then DrawImage3D Pressed\CheckBox, Gadget\X, Gadget\Y
	
	SetFont3D Style\Font[ GUI_T_CHECKBOX ]
	SetRGB Style\FontColor[ GUI_T_CHECKBOX ]
	
	Select Gadget\TextAlignment
		Case GUI_A_LEFT
			Text3D Gadget\X + ImageWidth3D( Style\Checkbox ) + 4, Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, False, True
		Case GUI_A_CENTER
			Text3D Gadget\X + Gadget\Width/2, Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, True, True
		Case GUI_A_RIGHT
			Text3D Gadget\X + Gadget\Width - StringWidth3D( Gadget\Caption$ ), Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, False, True
	End Select
End Function

Function RenderTab( Gadget.TGadget, Style.TGUIStyle )
	If Gadget\Pressed Then Style = Style\Pressed ElseIf Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	DrawGadgetRect Style, GUI_T_TAB, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	SetRGB Style\FontColor[ GUI_T_TAB ]
	SetFont3D Style\Font[ GUI_T_TAB ]
	
	If Gadget\Pressed Then
		OffsetX = 0
		OffsetY = -2
	EndIf
	
	Select Gadget\TextAlignment
		Case GUI_A_LEFT
			Text3D Gadget\X + 5 + OffsetX, Gadget\Y + Gadget\Height/2 + OffsetY, Gadget\Caption$, False, True
		Case GUI_A_CENTER
			Text3D Gadget\X + OffsetX + Gadget\Width/2, Gadget\Y + OffsetY + Gadget\Height/2, Gadget\Caption$, True, True
		Case GUI_A_RIGHT
			Text3D Gadget\X + Gadget\Width - StringWidth3D( Gadget\Caption$ ) - 5 + OffsetX, Gadget\Y + Gadget\Height/2 + OffsetY, Gadget\Caption$, False, True
	End Select
End Function

Function RenderRadioButton( Gadget.TGadget, Style.TGUIStyle )
	Pressed.TGUIStyle = Style\Pressed
	If Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	Color3D 255, 255, 255
	
	DrawImage3D Style\RadioCircle, Gadget\X, Gadget\Y
	
	If Gadget\Pressed Then DrawImage3D Pressed\RadioCircle, Gadget\X, Gadget\Y
	
	SetFont3D Style\Font[ GUI_T_RADIOBUTTON ]
	SetRGB Style\FontColor[ GUI_T_RADIOBUTTON ]
	
	Select Gadget\TextAlignment
		Case GUI_A_LEFT
			Text3D Gadget\X + ImageWidth3D( Style\Checkbox ) + 4, Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, False, True
		Case GUI_A_CENTER
			Text3D Gadget\X + Gadget\Width/2, Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, True, True
		Case GUI_A_RIGHT
			Text3D Gadget\X + Gadget\Width - StringWidth3D( Gadget\Caption$ ), Gadget\Y + ImageHeight3D( Style\Checkbox )/2, Gadget\Caption$, False, True
	End Select
End Function

Function RenderItemBox( Gadget.TGadget, Style.TGUIStyle )
	If Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	Color3D 255, 255, 255
	
	DrawGadgetRect Style, GUI_T_ITEMBOX, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	
	ItemBox.TItemBox = Object.TItemBox( Gadget\Group )
	
	X = ( Gadget\X - Itembox\X )/( Itembox\BoxSize + Itembox\Padding )
	Y = ( Gadget\Y - Itembox\Y )/( Itembox\BoxSize + Itembox\Padding )
	
	Image.TImageClass3D = Object.TImageClass3D( PeekInt( Itembox\IconBank, ( X + Y*Itembox\Width )*4 ) )
	
	If Image <> Null Then
		RectImage3D Image, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
		
		If PeekShort( Gadget\GadgetData, 0 ) > 1 Then
			SetRGB Style\FontColor[ GUI_T_ITEMBOX ]
			SetFont3D Style\Font[ GUI_T_ITEMBOX ]
			
			Select Gadget\TextAlignment
				Case GUI_A_LEFT
					Text3D Gadget\X + 2, Gadget\Y + Gadget\Height - FontHeight3D(), PeekShort( Gadget\GadgetData, 0 )
				Case GUI_A_CENTER
					Text3D Gadget\X + Gadget\Width/2, Gadget\Y + Gadget\Height - FontHeight3D(), PeekShort( Gadget\GadgetData, 0 ), True
				Case GUI_A_RIGHT
					Text3D Gadget\X + Gadget\Width - StringWidth3D( PeekShort( Gadget\GadgetData, 0 ) ) - 2, Gadget\Y + Gadget\Height - FontHeight3D(), PeekShort( Gadget\GadgetData, 0 )
			End Select
		EndIf
	EndIf
End Function

Function RenderComboBox( Gadget.TGadget, Style.TGUIStyle )
	If Gadget\Pressed Then Style = Style\Pressed ElseIf Gadget\MouseOver Then Style = Style\Hover Else Style = Style\Active
	
	Height = PeekShort( Gadget\GadgetData, 2 )
	
	Color3D 255, 255, 255
	
	DrawGadgetRect Style, GUI_T_COMBOBOX, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	If Gadget\Pressed Then
		DrawGadgetRect Style, GUI_T_COMBOBOX, Gadget\X, Gadget\Y, Gadget\Width, Height
		
		If Gadget\MouseOver Then
			MarkedEntry = ( MY - Gadget\Hwnd\Y - Gadget\Y - Gadget\Hwnd\TitlebarHeight )/Height
			
			If MarkedEntry Then RectImage3D Style\Highlight[ GUI_T_COMBOBOX ], Gadget\X, Gadget\Y + MarkedEntry*Height, Gadget\Width, Height, 0, True
		EndIf
	EndIf
	
	RectImage3D Style\Dropdown, Gadget\X + Gadget\Width - Height, Gadget\Y, Height, Height
	
	SetRGB Style\FontColor[ GUI_T_COMBOBOX ]
	SetFont3D Style\Font[ GUI_T_COMBOBOX ]
	
	If Gadget\Pressed Then
		Local Offset = 6
		For i = 1 To PeekShort( Gadget\GadgetData, 0 )
			Entry$ = PeekString( Gadget\GadgetData, Offset )
			
			Text3D Gadget\X + 5, Gadget\Y + i*Height + Height/2, Entry$, False, True
			
			Offset = Offset + Len( Entry$ ) + 1
		Next
	EndIf
	
	If PeekShort( Gadget\GadgetData, 4 ) Then
		Text3D Gadget\X + 5, Gadget\Y + Height/2, ComboboxMarkedEntry( Gadget ), False, True
	EndIf
End Function

Function RenderListbox( Gadget.TGadget, Style.TGUIStyle )
	Style = Style\Active
	
	DrawGadgetRect Style, GUI_T_LISTBOX, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	If PeekShort( Gadget\GadgetData, 2 ) Then
		RectImage3D Style\Highlight[ GUI_T_LISTBOX ], Gadget\X, Gadget\Y + ( PeekShort( Gadget\GadgetData, 2 ) - 1 )*PeekShort( Gadget\GadgetData, 4 ), Gadget\Width, PeekShort( Gadget\GadgetData, 4 ), 0, True
	EndIf
	
	SetRGB Style\FontColor[ GUI_T_LISTBOX ]
	SetFont3D Style\Font[ GUI_T_LISTBOX ]
	
	Local Offset = 6
	For i = 0 To PeekShort( Gadget\GadgetData, 0 ) - 1
		Entry$ = PeekString( Gadget\GadgetData, Offset )
		
		Text3D Gadget\X + 5, Gadget\Y + i*PeekShort( Gadget\GadgetData, 4 ), Entry$
		
		Offset = Offset + Len( Entry$ ) + 1
	Next
End Function

Function RenderTextfield( Gadget.TGadget, Style.TGUIStyle )
	Style = Style\Inactive
	
	DrawGadgetRect Style, GUI_T_TEXTFIELD, Gadget\X, Gadget\Y, Gadget\Width, Gadget\Height
	
	SetRGB Style\FontColor[ GUI_T_TEXTFIELD ]
	SetFont3D Style\Font[ GUI_T_TEXTFIELD ]
	
	Local CursorX
	Select Gadget\TextAlignment
		Case GUI_A_LEFT
			CursorX = Gadget\X + 5 + StringWidth3D( Gadget\Caption$ )
			Text3D Gadget\X + 5, Gadget\Y + Gadget\Height/2, Gadget\Caption$, False, True
		Case GUI_A_CENTER
			CursorX = Gadget\X + Gadget\Width/2 + StringWidth3D( Gadget\Caption$ )/2
			Text3D Gadget\X + Gadget\Width/2, Gadget\Y + Gadget\Height/2, Gadget\Caption$, True, True
		Case GUI_A_RIGHT
			CursorX = Gadget\X + Gadget\Width - 5
			Text3D Gadget\X + Gadget\Width - StringWidth3D( Gadget\Caption$ ) - 5, Gadget\Y + Gadget\Height/2, Gadget\Caption$, False, True
	End Select
	
	If Gadget\Active And ( ( MilliSecs() - PeekInt( Gadget\GadgetData, 0 ) ) Mod 1000 ) < 500 Then
		Line3D CursorX, Gadget\Y + 4, CursorX, Gadget\Y + Gadget\Height - 4
	EndIf
End Function

Function AddTab.TGadget( Tabber.TTabber, Width, Caption$ = "", TextAlignment = GUI_A_LEFT )
	X = Tabber\X
	For i = 0 To Tabber\TabCount - 1
		X = X + Tabber\Tabs[ i ]\Width
	Next
	
	Gadget.TGadget = New TGadget
		Gadget\Kind = GUI_T_TAB
		Gadget\Hwnd = Tabber\Hwnd
		Gadget\X = X
		Gadget\Y = Tabber\Y
		Gadget\Width = Width
		Gadget\Height = Tabber\TabHeight
		Gadget\Caption$ = Caption$
		Gadget\Group = Handle( Tabber )
		Gadget\TextAlignment = TextAlignment
		Gadget\Visible = True
	
	Tabber\Tabs[ Tabber\TabCount ] = Gadget
	Tabber\TabCount = Tabber\TabCount + 1
	
	ActivateTab( Gadget )
	
	Return Gadget
End Function


Function AddGadgetToTabber( Gadget.TGadget, Tab.TGadget )
	Gadget\TabGroup = Handle( Tab )
	Gadget\Visible = Tab\Pressed
	
	Tabber.TTabber = Object.TTabber( Tab\Group )
	
	Tabber\Gadgets[ Tabber\GadgetCount ] = Gadget
	Tabber\GadgetCount = Tabber\GadgetCount + 1
End Function

Function AddItemboxToTabber( Itembox.TItembox, Tab.TGadget )
	For i = 0 To BankSize( Itembox\GadgetBank ) - 4 Step 4
		AddGadgetToTabber( Object.TGadget( PeekInt( Itembox\GadgetBank, i ) ), Tab )
	Next
End Function

Function AddItemToItembox( Itembox.TItembox, X, Y, ItemHandle, ImagePath$, Name$ )
	Image.TImageClass3D = LoadImage3D( ImagePath$ )
	
	If Image = Null Then
		RuntimeError "Gib mal einen gltigen Bildpfad an, du unfhiges Ding"
	Else
		AddItemToBank( Itembox, X, Y, ItemHandle )
		PokeInt Itembox\IconBank, ( X + Y*Itembox\Width )*4, Handle( Image )
		
		Gadget.TGadget = Object.TGadget( PeekInt( Itembox\GadgetBank, ( X + Y*Itembox\Width )*4 ) )
			Gadget\Caption$ = Name$
	EndIf
End Function

Function AddItemToBank( Itembox.TItembox, X, Y, ItemHandle )
	Gadget.TGadget = Object.TGadget( PeekInt( Itembox\GadgetBank, ( X + Y*Itembox\Width )*4 ) )
	
	PokeShort Gadget\GadgetData, 0, PeekShort( Gadget\GadgetData, 0 ) + 1
	
	ResizeBank Gadget\GadgetData, BankSize( Gadget\GadgetData ) + 4
	PokeInt Gadget\GadgetData, BankSize( Gadget\GadgetData ) - 4, ItemHandle
End Function

Function AddStashToBank( SourceBank, DestBank )
	SourceCount = PeekShort( SourceBank, 0 )
	DestCount = PeekShort( DestBank, 0 )
	
	ResizeBank DestBank, ( SourceCount + DestCount )*4 + 2
	CopyBank SourceBank, 2, DestBank, DestCount*4 + 2, SourceCount*4
	PokeShort DestBank, 0, SourceCount + DestCount
	
	ResizeBank SourceBank, 2
	PokeShort SourceBank, 0, 0
End Function

Function AddComboboxEntry( Gadget.TGadget, Source$ )
	Size = BankSize( Gadget\GadgetData )
	
	ResizeBank Gadget\GadgetData, Size + Len( Source$ ) + 1
	
	PokeString Gadget\GadgetData, Size, Source$
	PokeShort Gadget\GadgetData, 0, PeekShort( Gadget\GadgetData, 0 ) + 1
	PokeShort Gadget\GadgetData, 4, PeekShort( Gadget\GadgetData, 0 )
End Function

Function AddListboxEntry( Gadget.TGadget, Source$ )
	Size = BankSize( Gadget\GadgetData )
	
	ResizeBank Gadget\GadgetData, Size + Len( Source$ ) + 1
	
	PokeString Gadget\GadgetData, Size, Source$
	PokeShort Gadget\GadgetData, 0, PeekShort( Gadget\GadgetData, 0 ) + 1
	PokeShort Gadget\GadgetData, 2, PeekShort( Gadget\GadgetData, 0 )
End Function

Function PickItem( Gadget.TGadget )
	ItemBox.TItemBox = Object.TItemBox( Gadget\Group )
	
	X = ( Gadget\X - Itembox\X )/( Itembox\BoxSize + Itembox\Padding )
	Y = ( Gadget\Y - Itembox\Y )/( Itembox\BoxSize + Itembox\Padding )
	
	g_GUI_PickedIconX = X
	g_GUI_PickedIconY = Y
	g_GUI_PickedItembox = Itembox
	g_GUI_PickedGadget = Gadget
	g_GUI_PickedIcon = Object.TImageClass3D( PeekInt( Itembox\IconBank, ( X + Y*Itembox\Width )*4 ) )
	PokeInt Itembox\IconBank, ( X + Y*Itembox\Width )*4, 0
	
	If g_GUI_PickedIcon <> Null Then
		OffsetX = ( MX - Itembox\X - Itembox\Hwnd\X ) Mod ( Itembox\BoxSize + Itembox\Padding )
		OffsetY = ( MY - Itembox\Y - Itembox\Hwnd\Y - Itembox\Hwnd\TitlebarHeight ) Mod ( Itembox\BoxSize + Itembox\Padding )
		HandleImage3D g_GUI_PickedIcon, OffsetX, OffsetY
	EndIf
End Function

Function DropItem()
	For Gadget.TGadget = Each TGadget
		If Gadget\Kind = GUI_T_ITEMBOX And Gadget\MouseOver Then
			
			ItemBox.TItemBox = Object.TItemBox( Gadget\Group )
			
			X = ( Gadget\X - Itembox\X )/( Itembox\BoxSize + Itembox\Padding )
			Y = ( Gadget\Y - Itembox\Y )/( Itembox\BoxSize + Itembox\Padding )
			
			HandleImage3D g_GUI_PickedIcon, 0, 0
			
			If Gadget <> g_GUI_PickedGadget Then TempIcon = PeekInt( Itembox\IconBank, ( X + Y*Itembox\Width )*4 ) Else TempIcon = Handle( g_GUI_PickedIcon )
			TempItem = Gadget\GadgetData
			TempName$ = Gadget\Caption$
			
			If Itembox <> g_GUI_PickedItembox Then Itembox_MoveItem( g_GUI_PickedGadget\GadgetData, g_GUI_PickedItembox, ItemBox )
			
			If Gadget\Caption$ = g_GUI_PickedGadget\Caption$ And Gadget <> g_GUI_PickedGadget Then
				AddStashToBank( g_GUI_PickedGadget\GadgetData, Gadget\GadgetData )
				
				g_GUI_PickedGadget\Caption$ = ""
				FreeImage3D g_GUI_PickedIcon
			Else
				PokeInt Itembox\IconBank, ( X + Y*Itembox\Width )*4, Handle( g_GUI_PickedIcon )
				Gadget\GadgetData = g_GUI_PickedGadget\GadgetData
				Gadget\Caption$ = g_GUI_PickedGadget\Caption$
				
				PokeInt g_GUI_PickedItembox\IconBank, ( g_GUI_PickedIconX + g_GUI_PickedIconY*g_GUI_PickedItembox\Width )*4, TempIcon
				g_GUI_PickedGadget\Caption$ = TempName$
				g_GUI_PickedGadget\GadgetData = TempItem
				g_GUI_PickedIcon = Null
				Exit
			EndIf
		EndIf
	Next
	
	If g_GUI_PickedIcon <> Null Then
		For Window.TWindow = Each TWindow
			If Window\MouseOver Then
				PokeInt g_GUI_PickedItembox\IconBank, ( g_GUI_PickedIconX + g_GUI_PickedIconY*g_GUI_PickedItembox\Width )*4, Handle( g_GUI_PickedIcon )
				HandleImage3D g_GUI_PickedIcon, 0, 0
				g_GUI_PickedIcon = Null
			EndIf
		Next
		
		If g_GUI_PickedIcon <> Null Then
			Itembox_DropItem( g_GUI_PickedGadget\GadgetData )
			g_GUI_PickedGadget\GadgetData = CreateBank( 2 )
			g_GUI_PickedIcon = Null
		EndIf
	EndIf
End Function

Function DrawGadgetRect( Style.TGUIStyle, GT, X, Y, Width, Height )
	Color3D 255, 255, 255
	
	RectImage3D Style\Background[ GT ], X + ImageWidth3D( Style\BorderL[ GT ] ), Y + ImageHeight3D( Style\BorderT[ GT ] ), Width - ImageWidth3D( Style\BorderL[ GT ] ) - ImageWidth3D( Style\BorderR[ GT ] ), Height - ImageHeight3D( Style\BorderT[ GT ] ) - ImageHeight3D( Style\BorderB[ GT ] )
	RectImage3D Style\BorderT[ GT ], X + ImageWidth3D( Style\EdgeTL[ GT ] ), Y, Width - ImageWidth3D( Style\EdgeTL[ GT ] ) - ImageWidth3D( Style\EdgeTR[ GT ] ), ImageHeight3D( Style\BorderT[ GT ] )
	RectImage3D Style\BorderL[ GT ], X, Y + ImageHeight3D( Style\EdgeTL[ GT ] ), ImageWidth3D( Style\BorderL[ GT ] ), Height - ImageHeight3D( Style\EdgeTL[ GT ] ) - ImageHeight3D( Style\EdgeBL[ GT ] )
	RectImage3D Style\BorderR[ GT ], X + Width, Y + ImageHeight3D( Style\EdgeTR[ GT ] ), ImageWidth3D( Style\BorderR[ GT ] ), Height - ImageHeight3D( Style\EdgeTR[ GT ] ) - ImageHeight3D( Style\EdgeBR[ GT ] )
	RectImage3D Style\BorderB[ GT ], X + ImageWidth3D( Style\EdgeBL[ GT ] ), Y + Height, Width - ImageWidth3D( Style\EdgeBL[ GT ] ) - ImageWidth3D( Style\EdgeBR[ GT ] ), ImageHeight3D( Style\BorderB[ GT ] )
	DrawImage3D Style\EdgeTL[ GT ], X, Y
	DrawImage3D Style\EdgeTR[ GT ], X + Width, Y
	DrawImage3D Style\EdgeBL[ GT ], X, Y + Height
	DrawImage3D Style\EdgeBR[ GT ], X + Width, Y + Height
End Function

Function MoveWindowToFront( W.TWindow )
	For Window.TWindow = Each TWindow
		Window\Active = False
	Next
	
	Insert W Before First TWindow
	W\Active = True
End Function

Function RedistributeLabel( Gadget.TGadget, Style.TGUIStyle )
	Local TextLine$[ 512 ], Counter, Caption$ = Gadget\Caption$
	
	SetFont3D Style\Active\Font[ GUI_T_LABEL ]
	
	While 1
		Offset = Instr( Caption$, " " )
		Offset2 = Instr( Caption$, Chr( 10 ) )
		If Offset2 = 0 Then Offset2 = 9999999
		
		If Offset2 < Offset Then
			TextLine[ Counter ] = TextLine[ Counter ] + Left( Caption$, Offset2 - 1 )
			Caption$ = Right( Caption$, Len( Caption$ ) - Offset2 )
			Counter = Counter + 1
		Else
			If Offset Then Part$ = Left( Caption$, Offset ) Else Part$ = Caption$
			If StringWidth3D( TextLine[ Counter ] ) + StringWidth3D( Part$ ) > Gadget\Width Then
				Counter = Counter + 1
			Else
				TextLine[ Counter ] = TextLine[ Counter ] + Part$
				Caption$ = Right( Caption$, Len( Caption$ ) - Offset )
			EndIf
			
			If Not Offset Then Exit
		EndIf
	Wend
	
	Gadget\Caption$ = ""
	For i = 0 To Counter
		Gadget\Caption$ = Gadget\Caption$ + TextLine[ i ] + Chr( 10 )
	Next
	
	Gadget\Caption$ = Left( Gadget\Caption$, Len( Gadget\Caption$ ) - 1 )
End Function

Function SetLabelCaption( Gadget.TGadget, Caption$ )
	Gadget\Caption$ = Caption$
	
	RedistributeLabel( Gadget, g_GUI_CurrentStyle )
End Function

Function ActivateTab( Gadget.TGadget )
	Tabber.TTabber = Object.TTabber( Gadget\Group )
	
	For i = 0 To Tabber\TabCount - 1
		Tabber\Tabs[ i ]\Pressed = False
	Next
	
	Gadget\Pressed = True
	
	For i = 0 To Tabber\GadgetCount - 1
		Tab.TGadget = Object.TGadget( Tabber\Gadgets[ i ]\TabGroup )
		
		Tabber\Gadgets[ i ]\Visible = Tab\Pressed
	Next
End Function

Function ActivateRadioButton( Gadget.TGadget )
	For G.TGadget = Each TGadget
		If G\Hwnd = Gadget\Hwnd And G\Kind = Gadget\Kind And G\Group = Gadget\Group Then G\Pressed = False
	Next
	
	Gadget\Pressed = True
End Function

Function ComboboxDropdown( Gadget.TGadget )
	If Not Gadget\Pressed Then
		Insert Gadget After Last TGadget
		
		Entries = PeekShort( Gadget\GadgetData, 0 )
		If Entries = 0 Then Entries = 1
		
		Gadget\Height = ( Entries + 1 )*PeekShort( Gadget\GadgetData, 2 )
		
		Gadget\Pressed = True
	Else
		MarkedEntry = ( MY - Gadget\Hwnd\Y - Gadget\Y - Gadget\Hwnd\TitlebarHeight )/PeekShort( Gadget\GadgetData, 2 )
		
		If MarkedEntry > 0 And PeekShort( Gadget\GadgetData, 0 ) <> 0 Then PokeShort Gadget\GadgetData, 4, MarkedEntry
		
		ComboboxDropup( Gadget )
	EndIf
End Function

Function ComboboxSelectEntry( Gadget.TGadget, Index )
	PokeShort( Gadget\GadgetData, 4, Index )
End Function

Function ComboboxDropup( Gadget.TGadget )
	Gadget\Height = PeekShort( Gadget\GadgetData, 2 )
	Gadget\Pressed = False
End Function

Function ComboboxMarkedEntryIndex( Gadget.TGadget )
	Return PeekShort( Gadget\GadgetData, 4 )
End Function

Function ComboboxMarkedEntry$( Gadget.TGadget )
	Offset = 6
	For i = 1 To PeekShort( Gadget\GadgetData, 4 )
		Entry$ = PeekString$( Gadget\GadgetData, Offset )
		Offset = Offset + Len( Entry$ ) + 1
	Next
	
	Return Entry$
End Function

Function ListboxPickEntry( Gadget.TGadget )
	SelectedEntry = ( MY - Gadget\Hwnd\Y - Gadget\Y - Gadget\Hwnd\TitlebarHeight )/PeekShort( Gadget\GadgetData, 4 )
	
	ListboxMarkEntry( Gadget, SelectedEntry + 1 )
End Function

Function ListboxMarkEntry( Gadget.TGadget, Index )
	If Index <= PeekShort( Gadget\GadgetData, 0 ) Then
		PokeShort Gadget\GadgetData, 2, Index
	Else
		PokeShort Gadget\GadgetData, 2, PeekShort( Gadget\GadgetData, 0 )
	EndIf
End Function

Function ListboxMarkedEntryIndex( Gadget.TGadget )
	Return PeekShort( Gadget\GadgetData, 2 )
End Function

Function ListboxMarkedEntry$( Gadget.TGadget )
	Offset = 6
	For i = 1 To PeekShort( Gadget\GadgetData, 2 )
		Entry$ = PeekString$( Gadget\GadgetData, Offset )
		Offset = Offset + Len( Entry$ ) + 1
	Next
	
	Return Entry$
End Function

Function ListboxClear( Gadget.TGadget )
	PokeInt Gadget\GadgetData, 0, 0
	ResizeBank Gadget\GadgetData, 6
End Function

Function LoadStyle.TGUIStyle( Path$ )
	Size = FileSize( Path$ )
	Instream = ReadFile( Path$ )
	
	If Instream Then
		Style.TGUIStyle = New TGUIStyle
			Style\Active = New TGUIStyle
			Style\Inactive = New TGUIStyle
			Style\Hover = New TGUIStyle
			Style\Pressed = New TGUIStyle
			Style\Tabber = New TGUIStyle
		
		SeekFile Instream, Size - 8
		TableOffset = ReadInt( Instream )
		ImageOffset = ReadInt( Instream )
		
		SeekFile Instream, TableOffset
		Entries = ReadShort( Instream )
		LookupTable = CreateBank( Entries*4 )
		ReadBytes LookupTable, Instream, 0, Entries*4
		
		SeekFile Instream, ImageOffset
		ContentLength = ReadInt( Instream )
		ImageBank = CreateBank( ContentLength )
		ReadBytes ImageBank, Instream, 0, ContentLength
		
		SeekFile Instream, 0
		
		While Not Eof( Instream )
			Chunk = ReadByte( Instream )
			
			Select Chunk
				Case D_END_STYLE
					Exit
				Case D_WINDOW
					QuitChunk = D_END_WINDOW
					GadgetType = GUI_T_WINDOW
				Case D_LABEL
					QuitChunk = D_END_LABEL
					GadgetType = GUI_T_LABEL
				Case D_BUTTON
					QuitChunk = D_END_BUTTON
					GadgetType = GUI_T_BUTTON
				Case D_CHECKBOX
					QuitChunk = D_END_CHECKBOX
					GadgetType = GUI_T_CHECKBOX
				Case D_TAB
					QuitChunk = D_END_TAB
					GadgetType = GUI_T_TAB
				Case D_RADIOBUTTON
					QuitChunk = D_END_RADIOBUTTON
					GadgetType = GUI_T_RADIOBUTTON
				Case D_ITEMBOX
					QuitChunk = D_END_ITEMBOX
					GadgetType = GUI_T_ITEMBOX
				Case D_LISTBOX
					QuitChunk = D_END_LISTBOX
					GadgetType = GUI_T_LISTBOX
				Case D_COMBOBOX
					QuitChunk = D_END_COMBOBOX
					GadgetType = GUI_T_COMBOBOX
				Case D_TEXTFIELD
					QuitChunk = D_END_TEXTFIELD
					GadgetType = GUI_T_TEXTFIELD
			End Select
			
			If QuitChunk Then
				While Chunk <> QuitChunk
					Chunk = ReadByte( Instream )
					
					Select Chunk
						Case D_ACTIVE
							ExitChunk = D_END_ACTIVE
							LoadStyle.TGUIStyle = Style\Active
						Case D_INACTIVE
							ExitChunk = D_END_INACTIVE
							LoadStyle.TGUIStyle = Style\Inactive
						Case D_HOVER
							ExitChunk = D_END_HOVER
							LoadStyle.TGUIStyle = Style\Hover
						Case D_PRESSED
							ExitChunk = D_END_PRESSED
							LoadStyle.TGUIStyle = Style\Pressed
						Case D_TABBER
							ExitChunk = D_END_TABBER
							LoadStyle.TGUIStyle = Style\Tabber
					End Select
					
					If ExitChunk Then
						Chunk = ReadByte( Instream )
						
						While Chunk <> ExitChunk
							ParseIniCommand( Chunk, Instream, LoadStyle, GadgetType )
							Chunk = ReadByte( Instream )
						Wend
						
						ExitChunk = 0
					EndIf
				Wend
				
				QuitChunk = 0
			EndIf
		Wend
		
		Return Style
	Else
		RuntimeError "Style doesn't exist: " + Path$
	EndIf
End Function

Function ParseIniCommand( Chunk, Instream, Style.TGUIStyle, GadgetType )
	Select Chunk
		Case D_EDGE_TL
			Style\EdgeTL[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_EDGE_TR
			Style\EdgeTR[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
			HandleImage3D Style\EdgeTR[ GadgetType ], ImageWidth3D( Style\EdgeTR[ GadgetType ] ), 0
		Case D_EDGE_BL
			Style\EdgeBL[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
			HandleImage3D Style\EdgeBL[ GadgetType ], 0, ImageHeight3D( Style\EdgeBL[ GadgetType ] )
		Case D_EDGE_BR
			Style\EdgeBR[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
			HandleImage3D Style\EdgeBR[ GadgetType ], ImageWidth3D( Style\EdgeBR[ GadgetType ] ), ImageHeight3D( Style\EdgeBR[ GadgetType ] )
		Case D_BORDER_T
			Style\BorderT[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_BORDER_B
			Style\BorderB[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
			HandleImage3D Style\BorderB[ GadgetType ], 0, ImageHeight3D( Style\BorderB[ GadgetType ] )
		Case D_BORDER_L
			Style\BorderL[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_BORDER_R
			Style\BorderR[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
			HandleImage3D Style\BorderR[ GadgetType ], ImageWidth3D( Style\BorderR[ GadgetType ] ), 0
		Case D_FONTSIZE
			Style\Fontsize[ GadgetType ] = ReadShort( Instream )
		Case D_FONT
			Style\Font[ GadgetType ] = LoadFontFromStyle( ReadShort( Instream ), Style\Fontsize[ GadgetType ] )
		Case D_FONTCOLOR
			Style\FontColor[ GadgetType ] = ReadInt( Instream )
		Case D_BACKGROUND
			Style\Background[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_BOX
			Style\CheckBox = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_CIRCLE
			Style\RadioCircle = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_DROPDOWN
			Style\Dropdown = LoadImageFromStyle( ReadShort( Instream ) )
		Case D_HIGHLIGHT
			Style\Highlight[ GadgetType ] = LoadImageFromStyle( ReadShort( Instream ) )
		Default
			RuntimeError "Something wrong here. Style corrupt?"
	End Select
End Function

Function LoadImageFromStyle.TImageClass3D( ID )
	If Images( ID ) <> Null Then Return Images( ID )
	
	Offset = PeekInt( LookupTable, ID*4 )
	
	Width = PeekShort( ImageBank, Offset + 0 )
	Height = PeekShort( ImageBank, Offset + 2 )
	Offset = Offset + 4
	
	Texture = CreateTexture( 2^Ceil( Log( Width )/Log( 2 ) ), 2^Ceil( Log( Height )/Log( 2 ) ), 4 )
	Buffer = TextureBuffer( Texture )
	
	LockBuffer Buffer
	For X = 0 To Width - 1
		For Y = 0 To Height - 1
			RGB = PeekByte( ImageBank, Offset ) Shl 16 + PeekByte( ImageBank, Offset + 1 ) Shl 8 + PeekByte( ImageBank, Offset + 2 )
			
			If RGB Then
				WritePixelFast X, Y, $FF000000 + RGB, Buffer
			Else
				WritePixelFast X, Y, 0, Buffer
			EndIf
			Offset = Offset + 3
		Next
	Next
	UnlockBuffer Buffer
	
	Images( ID ) = LoadImage3D( "", False, 4, Texture )
	
	Images( ID )\FWidth = Width
	Images( ID )\FHeight = Height
	Images( ID )\Animated = True
	
	Return Images( ID )
End Function

Function LoadFontFromStyle.TFont3D( ID, Size )
	If Fonts( ID ) <> Null Then Return Fonts( ID )
	
	Offset = PeekInt( LookupTable, ID*4 )
	
	Width = PeekShort( ImageBank, Offset + 0 )*8
	Height = PeekShort( ImageBank, Offset + 2 )
	Offset = Offset + 4
	
	Texture = CreateTexture( 2^Ceil( Log( Width )/Log( 2 ) ), 2^Ceil( Log( Height )/Log( 2 ) ), 4 )
	Buffer = TextureBuffer( Texture )
	
	LockBuffer Buffer
	For Y = 0 To Height - 1
		For X = 0 To Width - 1
			Byte = X/8
			Bit = X Mod 8
			
			Filled = Sgn( PeekByte( ImageBank, Offset + Byte ) And ( 1 Shl Bit ) )
			WritePixelFast X, Y, Filled*$FFFFFFFF, Buffer
		Next
		
		Offset = Offset + Width/8
	Next
	UnlockBuffer Buffer
	
	Fonts( ID ) = LoadFont3D( "", Size, False, False, False, False, 4, Texture )
	
	Return Fonts( ID )
End Function

Function SetRGB( RGB, Alpha# = 1 )
	Color3D ( RGB Shr 16 ) And $FF, ( RGB Shr 8 ) And $FF, RGB And $FF, Alpha#
End Function

Function PeekString$( Bank, Offset )
	Local Outstring$ = ""
	
	Char = PeekByte( Bank, Offset )
	While Char
		Outstring$ = Outstring$ + Chr( Char )
		
		Offset = Offset + 1
		Char = PeekByte( Bank, Offset )
	Wend
	
	Return Outstring$
End Function

Function PokeString( Bank, Offset, Source$ )
	For i = 0 To Len( Source$ ) - 1
		PokeByte Bank, Offset + i, Asc( Mid( Source$, i + 1, 1 ) )
	Next
	
	PokeByte Bank, Offset + Len( Source$ ), 0
End Function