Global BlockHUDImage, CursorHUDImage.TImageClass3D, MinimapHUDImage
Global MinimapHUDX, MinimapHUDY, MinimapHUDWidth, MinimapHUDHeight, MinimapFullScreen = False

Function DrawHUD()
	RenderPlayerNames()
	Render3DCursor()
	RenderBlocks()
	RenderMinimap()
End Function

Function RenderBlocks()
	If MinimapFullscreen Then Return
	
	Local BlocksPerScreen = Floor( ( Config\GWidth - MinimapHUDWidth )/Float( TEXTURE_SCREEN_SIZE ) )
	
	If BlocksPerScreen > BlockCounter - 1 Then BlocksPerScreen = BlockCounter - 1
	
	Origin ( Config\GWidth - MinimapHUDWidth )/2 - BlocksPerScreen*TEXTURE_SCREEN_SIZE/2, Config\GHeight - TEXTURE_SCREEN_SIZE
	
	Color 0, 0, 0
	Rect -2, -2, BlocksPerScreen*TEXTURE_SCREEN_SIZE + 4, TEXTURE_SCREEN_SIZE + 4
	
	Local StartBlock = SelectedBlock - BlocksPerScreen/2
	
	If StartBlock < 1 Then StartBlock = 1
	If StartBlock + BlocksPerScreen > BlockCounter Then StartBlock = BlockCounter - BlocksPerScreen
	
	Color 255, 255, 255
	For I = 0 To BlocksPerScreen - 1
		Local X = ( ( I + StartBlock ) Mod MAX_BLOCK_COUNT_SQR )*TEXTURE_SCREEN_SIZE
		Local Y = ( ( I + StartBlock )  /  MAX_BLOCK_COUNT_SQR )*TEXTURE_SCREEN_SIZE
		DrawImageRect BlockHUDImage, I*TEXTURE_SCREEN_SIZE, 0, X, Y, TEXTURE_SCREEN_SIZE, TEXTURE_SCREEN_SIZE
	Next
	
	Color 191, 191, 191
	Rect ( SelectedBlock - StartBlock )*TEXTURE_SCREEN_SIZE, 0, TEXTURE_SCREEN_SIZE, TEXTURE_SCREEN_SIZE, False
	
	Color 255, 255, 255
	
	Origin 0, 0
End Function

Function PreRenderBlocks()
	If ( Not FileType( "Models\PrerenderedBlocks.png" ) ) And ( Not FileType( "Models\PrerenderedBlocks.bmp" ) ) Then
		Local Image = CreateImage( MAX_BLOCK_COUNT_SQR*TEXTURE_SCREEN_SIZE, MAX_BLOCK_COUNT_SQR*TEXTURE_SCREEN_SIZE )
		
		MapW = 3
		MapH = 3
		MapD = 3
		InitMap()
		
		Local Pivot = CreatePivot()
		PositionEntity Pivot, 1.0, 1.0, 1.001
		
		Local Cam = CreateCamera( Pivot )
		CameraClsColor Cam, 64, 64, 64
		CameraProjMode Cam, 2
		PositionEntity Cam, EntityX( Pivot ), EntityY( Pivot ), EntityZ( Pivot ), True
		MoveEntity Cam, 0.0, 0.0, -1.9
		TurnEntity Pivot, 30.0, 0.0, 0.0, True
		TurnEntity Pivot, 0.0, 45.0, 0.0, True
		CameraViewport Cam, 0, 0, TEXTURE_SCREEN_SIZE, TEXTURE_SCREEN_SIZE
		
		InitDX7Hack()
		DisableTextureFilters()
		
		For I = 0 To MAX_BLOCK_COUNT - 1
			If BlockTypes[ I ] <> Null Then
				If BlockTypes[ I ]\IsObject Then
					Local GameObject.TGameObject = FindOrCreateGameObject( 1, 1, 1 )
					
					GameObject\Kind = I
					GameObject\NZ# = -1
				EndIf
				
				MapData( 1, 1, 1 ) = I
				
				BuildChunk( 0, 0, 0 )
				
				RenderWorld
				
				Local X = ( I Mod MAX_BLOCK_COUNT_SQR )*TEXTURE_SCREEN_SIZE
				Local Y = ( I  /  MAX_BLOCK_COUNT_SQR )*TEXTURE_SCREEN_SIZE
				CopyRect 0, 0, TEXTURE_SCREEN_SIZE, TEXTURE_SCREEN_SIZE, X, Y, BackBuffer(), ImageBuffer( Image )
			EndIf
		Next
		
		FreeEntity Cam
		FreeEntity Pivot
		
		SaveImage( Image, "Models\PrerenderedBlocks.bmp" )
		
		End
	EndIf
	
	If FileType( "Models\PrerenderedBlocks.png" ) Then
		BlockHUDImage = LoadImage( "Models\PrerenderedBlocks.png" )
	ElseIf FileType( "Models\PrerenderedBlocks.bmp" ) Then
		BlockHUDImage = LoadImage( "Models\PrerenderedBlocks.bmp" )
	EndIf
	
	MapW = Config\MapWidth
	MapH = Config\MapHeight
	MapD = Config\MapDepth
End Function

Function Render3DCursor()
	If Not Config\CursorEnabled Then Return
	
	Local Mesh = CameraPick( LocalPlayer\Camera, MouseX(), MouseY() )
	
	If Mesh Then
		Local Surface = PickedSurface()
		Local Vertex = TriangleVertex( Surface, PickedTriangle(), 0 )
		
		TFormPoint PickedX(), PickedY(), PickedZ(), 0, LocalPlayer\Entity
		
		If TFormedZ() < MAX_BLOCK_DISTANCE Then
			Local DirX# = -Sgn( PickedNX() )*0.1
			Local DirY# = -Sgn( PickedNY() )*0.1
			Local DirZ# = -Sgn( PickedNZ() )*0.1
			
			Local VoxelX = Int( PickedX() + DirX# )
			Local VoxelY = Int( PickedY() + DirY# )
			Local VoxelZ = Int( PickedZ() + DirZ# )
			
			Color3D 128, 128, 128
			If MapGOs( VoxelX, VoxelY, VoxelZ ) <> Null Then
				Local GO.TGameObject = MapGOs( VoxelX, VoxelY, VoxelZ )
				Vertex = GO\PickIndex
				
				For I = 0 To 3
					OutlineVertices( Mesh, Surface, Vertex + I,     Vertex + ( ( I + 1 ) Mod 4 )     )
					OutlineVertices( Mesh, Surface, Vertex + I + 4, Vertex + ( ( I + 1 ) Mod 4 ) + 4 )
					OutlineVertices( Mesh, Surface, Vertex + I,     Vertex + I + 4 )
				Next
			Else
				Vertex = ( Vertex/4 )*4
				
				For I = 0 To 3
					OutlineVertices( Mesh, Surface, Vertex + I, Vertex + ( ( I + 1 ) Mod 4 ) )
				Next
			EndIf
		EndIf
	EndIf
End Function

Function RenderPlayerNames()
	For Player.TPlayer = Each TPlayer
		If Player <> LocalPlayer Then
			CameraProject LocalPlayer\Camera, Player\X#, Player\Y# + 1.5, Player\Z#
			TFormPoint Player\X#, Player\Y#, Player\Z#, 0, LocalPlayer\Entity
			
			If ProjectedX() <> 0 Or ProjectedY() <> 0 Then
				Local Scale# = ( 1.0 - TFormedZ()/VisibilityRange )^5
				
				If Scale# < 0.0 Then Scale# = 0.0 ElseIf Scale# >= 1.0 Then Scale# = 1.0
				
				Color3D 255, 255, 255
				ScaleImage3D CurrFont\ImageClass, Scale#*2.0, Scale#*2.0
				Text3D ProjectedX(), ProjectedY(), Player\Name$, True, True
				ScaleImage3D CurrFont\ImageClass, 1.0, 1.0
			EndIf
		EndIf
	Next
End Function

Function LoadCursor()
	CursorHUDImage = LoadImage3D( "Models\Cursor.png" )
	MaskImage3D CursorHUDImage, 1, 1, 1
End Function

Function DrawCursor()
	If Config\FullScreen Then DrawImage3D CursorHUDImage, MouseX(), MouseY()
End Function

Function OutlineVertices( Mesh, Surface, V1, V2 )
	TFormPoint VertexX( Surface, V1 ), VertexY( Surface, V1 ), VertexZ( Surface, V1 ), Mesh, 0
	CameraProject( LocalPlayer\Camera, TFormedX(), TFormedY(), TFormedZ() )
	
	ProjX = ProjectedX()
	ProjY = ProjectedY()
	
	TFormPoint VertexX( Surface, V2 ), VertexY( Surface, V2 ), VertexZ( Surface, V2 ), Mesh, 0
	CameraProject( LocalPlayer\Camera, TFormedX(), TFormedY(), TFormedZ() )
	
	If Not ( ProjX <> 0 And ProjY <> 0 And ProjectedX() <> 0 And ProjectedY() <> 0 ) Then Return
	
	Line3D ProjX, ProjY, ProjectedX(), ProjectedY()
End Function

Function ToggleFullscreenMinimap()
	MinimapFullscreen = Not MinimapFullscreen
	
	If MinimapFullscreen Then
		MinimapHUDWidth  = Config\GWidth
		MinimapHUDHeight = Config\GHeight
		MinimapHUDX = 0
		MinimapHUDY = 0
	Else
		MinimapHUDWidth  = 200
		MinimapHUDHeight = 200
		MinimapHUDX = Config\GWidth - MinimapHUDWidth
		MinimapHUDY = Config\GHeight - MinimapHUDHeight
	EndIf
End Function

Function RenderMinimap()
	Origin MinimapHUDX, MinimapHUDY
	
	ProjectVoxelToMinimap( LocalPlayer\X#, LocalPlayer\Y#, LocalPlayer\Z# )
	
	Local CenterX = MinimapProjectedX
	Local CenterY = MinimapProjectedY
	
	If Not MinimapFullScreen Then
		Color 0, 0, 0
		Rect -3, -3, MinimapHUDWidth + 3, 3
		Rect -3, -3, 3, MinimapHUDHeight + 3
		Color 128, 128, 128
		Rect -2, -2, MinimapHUDWidth + 2, 2
		Rect -2, -2, 2, MinimapHUDHeight + 2
		Color 255, 255, 255
		Rect 0, 0, MinimapHUDWidth, MinimapHUDHeight
	EndIf
	
	;DrawImage MinimapHUDImage, 0, 0
	DrawImageRect MinimapHUDImage, 0, 0, CenterX - MinimapHUDWidth/2, CenterY - MinimapHUDHeight/2, MinimapHUDWidth, MinimapHUDHeight
	
	For Player.TPlayer = Each TPlayer
		ProjectVoxelToMinimap( Player\X#, Player\Y#, Player\Z# )
		
		If RectsOverlap( CenterX - MinimapHUDWidth/2, CenterY - MinimapHUDHeight/2, MinimapHUDWidth, MinimapHUDHeight, MinimapProjectedX, MinimapProjectedY, 1, 1 ) Then
			Color 255, 255, 0
			
			Rect MinimapProjectedX - CenterX + MinimapHUDWidth/2 - 2, MinimapProjectedY - CenterY + MinimapHUDHeight/2 - 2, 4, 4
		EndIf
	Next
	
	Origin 0, 0
End Function