Global LuaState

Global BlockTypeCounter = 1, BlockCounter

Function InitLua()
	LuaState = LuaL_NewState()
	LuaL_OpenLibs( LuaState )
	Lua_InitInterface( LuaState, "CallFunction" )
	
	InitEmptyBlock()
End Function

Function LoadBlocks()
	Lua_NewTable( LuaState )
	Lua_SetGlobal( LuaState, "Blocks" )
	
	Local Status = LuaL_LoadFile( LuaState, "Data/Blocks.lua" )
	If Status Then RuntimeError "Error while loading lua script:" + Chr( 10 ) + Lua_ToString( LuaState, 1 )
	
	Status = Lua_PCall( LuaState, 0, 0, 0 )
	If Status Then RuntimeError "Error while executing lua script:" + Chr( 10 ) + Lua_ToString( LuaState, 1 )
	
	Lua_GetGlobal( LuaState, "Blocks" )
	Local Index = Lua_GetTop( LuaState )
	
	Lua_PushNil( LuaState )
	While Lua_Next( LuaState, Index )
		LoadBlock()
		
		Lua_Pop( LuaState, 1 )
	Wend
	Lua_Pop( LuaState, 1 )
	
	While BlockTypes[ BlockCounter ] <> Null
		BlockCounter = BlockCounter + 1
	Wend
End Function

Function InitEmptyBlock()
	Local Block.TBlockType = New TBlockType
	
	Block\Transparent = True
	Block\TransIndex = -9999
	
	BlockTypes[ 0 ] = Block
End Function

Function LoadBlock()
	Local Block.TBlockType = New TBlockType
	
	Local Index = Lua_GetTop( LuaState )
	
	Lua_GetField( LuaState, Index, "Faces" )
	Lua_GetField( LuaState, Index, "Transparent" )
	Lua_GetField( LuaState, Index, "Colliding" )
	Lua_GetField( LuaState, Index, "Flowing" )
	Lua_GetField( LuaState, Index, "Transparency" )
	Lua_GetField( LuaState, Index, "TransparentIndex" )
	Lua_GetField( LuaState, Index, "ID" )
	Lua_GetField( LuaState, Index, "DisplayImage" )
	Lua_GetField( LuaState, Index, "ForceBorders" )
	Lua_GetField( LuaState, Index, "Gravity" )
	Lua_GetField( LuaState, Index, "IsObject" )
	Lua_GetField( LuaState, Index, "ModelPath" )
	Lua_GetField( LuaState, Index, "IsLight" )
	Lua_GetField( LuaState, Index, "LightRadius" )
	Lua_GetField( LuaState, Index, "DestroyImage" )
	Lua_GetField( LuaState, Index, "IsTrigger" )
	Lua_GetField( LuaState, Index, "TriggeredOnTouch" )
	Lua_GetField( LuaState, Index, "TriggeredBlock" )
	Lua_GetField( LuaState, Index, "ReplaceWith" )
	Lua_GetField( LuaState, Index, "TriggeredOnAction" )
	Lua_GetField( LuaState, Index, "ReplaceBlocks" )
	Lua_GetField( LuaState, Index, "AdjustToNormal" )
	Lua_GetField( LuaState, Index, "LightR" )
	Lua_GetField( LuaState, Index, "LightG" )
	Lua_GetField( LuaState, Index, "LightB" )
	Lua_GetField( LuaState, Index, "Masked" )
	Lua_GetField( LuaState, Index, "MinimapColorR" )
	Lua_GetField( LuaState, Index, "MinimapColorG" )
	Lua_GetField( LuaState, Index, "MinimapColorB" )
	
	Block\IsObject  = Lua_ToBoolean( LuaState, Index + 11 )
	Block\IsLight   = Lua_ToBoolean( LuaState, Index + 13 )
	Block\IsTrigger = Lua_ToBoolean( LuaState, Index + 16 )
	
	If Block\IsObject Then
		Block\AdjustToNormal = Lua_ToBoolean( LuaState, Index + 22 )
		Block\Model = LoadModelData( Lua_ToString( LuaState, Index + 12 ), Block\AdjustToNormal )
	Else
		Local FacesIndex = Lua_GetTop( LuaState )
		Lua_GetField( LuaState, Index + 1, "Left" )
		Lua_GetField( LuaState, Index + 1, "Right" )
		Lua_GetField( LuaState, Index + 1, "Front" )
		Lua_GetField( LuaState, Index + 1, "Back" )
		Lua_GetField( LuaState, Index + 1, "Top" )
		Lua_GetField( LuaState, Index + 1, "Bottom" )
		
		For I = 1 To 6
			Block\FaceTexture[ I - 1 ] = Lua_ToNumber( LuaState, FacesIndex + I )
		Next
	EndIf
	
	If Block\IsLight Then
		Block\LightRadius# = Lua_ToNumber( LuaState, Index + 14 )
		Block\InvLightRadius# = 1.0/Block\LightRadius#
		Block\LightR# = Lua_ToNumber( LuaState, Index + 23 )
		Block\LightG# = Lua_ToNumber( LuaState, Index + 24 )
		Block\LightB# = Lua_ToNumber( LuaState, Index + 25 )
	EndIf
	
	If Block\IsTrigger Then
		Block\TriggeredOnTouch = Lua_ToBoolean( LuaState, Index + 17 )
		Block\TriggeredBlock = Lua_ToNumber( LuaState, Index + 18 )
		Block\ReplaceWith = Lua_ToNumber( LuaState, Index + 19 )
		Block\TriggeredOnAction = Lua_ToBoolean( LuaState, Index + 20 )
		
		If Not Lua_IsNil( LuaState, Index + 21 ) Then
			Local Length = Lua_ObjLen( LuaState, Index + 21 )
			
			If Length Mod 2 Then RuntimeError( "Lua Error: ReplaceBlocks must have even element count" )
			
			Block\ReplaceBlockCount = Length
			
			Local BlocksIndex = Lua_GetTop( LuaState )
			For I = 0 To Length - 1
				Lua_RawGetI( LuaState, Index + 21, I + 1 )
				Block\ReplaceBlocks[ I ] = Lua_ToNumber( LuaState, BlocksIndex + I + 1 )
			Next
		EndIf
	EndIf
	
	Block\Transparent = Lua_ToBoolean( LuaState, Index + 2 )
	Block\Colliding = Lua_ToBoolean( LuaState, Index + 3 )
	Block\Flowing = Lua_ToBoolean( LuaState, Index + 4 )
	Block\Masked = Lua_ToBoolean( LuaState, Index + 26 )
	
	If Block\Masked Then Block\TransIndex = -1
	
	If Block\Transparent Or Block\Masked Then
		If Not Lua_IsNil( LuaState, Index + 5 ) Then Block\Transparency = Lua_ToNumber( LuaState, Index + 5 )*0.5
		If Not Lua_IsNil( LuaState, Index + 6 ) Then Block\TransIndex = -Lua_ToNumber( LuaState, Index + 6 ) - 1
	EndIf
	
	Block\Kind = Lua_ToNumber( LuaState, Index + 7 )
	Block\DisplayImage = Lua_ToNumber( LuaState, Index + 8 )
	Block\DestroyImage = Lua_ToNumber( LuaState, Index + 15 )
	Block\ForceBorders = Lua_ToBoolean( LuaState, Index + 9 )
	Block\Gravity = Lua_ToNumber( LuaState, Index + 10 )
	
	Block\MinimapR = Lua_ToNumber( LuaState, Index + 27 )
	Block\MinimapG = Lua_ToNumber( LuaState, Index + 28 )
	Block\MinimapB = Lua_ToNumber( LuaState, Index + 29 )
	
	Lua_SetTop( LuaState, Index )
	
	If BlockTypes[ Block\Kind ] = Null Then
		BlockTypes[ Block\Kind ] = Block
	Else
		RuntimeError( "Lua Error: Duplicate block ID '" + Block\Kind + "'" )
	EndIf
	
	Block\IsSolidBlock = ( Not Block\Transparent ) And ( Not Block\IsObject ) And ( Not Block\Masked )
	
	BlockTypeCounter = BlockTypeCounter + 1
End Function

Function Lua_UserFunction()
	L = Lua_GetCurrentState()
	
	FuncName$ = Lua_ToString( L, 1 )
	
	Select FuncName$
	End Select
End Function

Function LoadModelData.TModelData( Path$, CreateRotations )
	LoaderMatrix( "3ds", 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 )
	
	Local Mesh = LoadMesh( CurrentDir() + Path$ )
	
	If Mesh Then
		UpdateNormals Mesh
		
		Local ModelData.TModelData = LoadModelDataFromMesh( Mesh )
		
		If CreateRotations Then
			For I = 0 To 2
				RotateMesh Mesh, 0, 90, 0
				ModelData\Rotations[ I ] = LoadModelDataFromMesh( Mesh )
			Next
		EndIf
		
		FreeEntity Mesh
		
		Return ModelData
	Else
		RuntimeError( "Unable to load model with path " + CurrentDir() + Path$ )
	EndIf
End Function

Function LoadModelDataFromMesh.TModelData( Mesh )
	Local ModelData.TModelData = New TModelData

	Local Surface = GetSurface( Mesh, 1 )
	
	ModelData\VertexCount = CountVertices( Surface )
	ModelData\TriangleCount = CountTriangles( Surface )
	
	If ModelData\VertexCount > 64 Or ModelData\TriangleCount > 64 Then RuntimeError( "Model exceeding allowed polygon count at path " + CurrentDir() + Path$ )
	
	ModelData\MinX# = 100.0
	ModelData\MinY# = 100.0
	ModelData\MinZ# = 100.0
	
	For I = 0 To ModelData\VertexCount - 1
		ModelData\X[ I ] = VertexX( Surface, I )
		ModelData\Y[ I ] = VertexY( Surface, I )
		ModelData\Z[ I ] = VertexZ( Surface, I )
		ModelData\NX[ I ] = VertexNX( Surface, I )
		ModelData\NY[ I ] = VertexNY( Surface, I )
		ModelData\NZ[ I ] = VertexNZ( Surface, I )
		ModelData\U[ I ] = VertexU( Surface, I )
		ModelData\V[ I ] = VertexV( Surface, I )
		
		If ModelData\X[ I ] < ModelData\MinX# Then ModelData\MinX# = ModelData\X[ I ]
		If ModelData\Y[ I ] < ModelData\MinY# Then ModelData\MinY# = ModelData\Y[ I ]
		If ModelData\Z[ I ] < ModelData\MinZ# Then ModelData\MinZ# = ModelData\Z[ I ]
	Next
	
	ModelData\Width# = MeshWidth( Mesh )
	ModelData\Height# = MeshHeight( Mesh )
	ModelData\Depth# = MeshDepth( Mesh )
	
	For I = 0 To ModelData\TriangleCount - 1
		ModelData\T[ I*3     ] = TriangleVertex( Surface, I, 0 )
		ModelData\T[ I*3 + 1 ] = TriangleVertex( Surface, I, 1 )
		ModelData\T[ I*3 + 2 ] = TriangleVertex( Surface, I, 2 )
	Next
	
	Return ModelData
End Function