Const MINIMAP_HUD_SIZE = 128
Const MINIMAP_SCALE# = 2.0

Global DDA_X, DDA_Y, DDA_Z, DDA_NX, DDA_NY, DDA_NZ, DDA_ThroughWater

Global MinimapWidth, MinimapHeight
Global MinimapRenderX, MinimapRenderY, MinimapRenderDone
Global MinimapProjectedX, MinimapProjectedY

Function PreRenderMinimap()
	Local Width  = ( Cos( 45 )*MapW + Sin( 45 )*MapD )*MINIMAP_SCALE#
	Local Height = ( Sin( 45 )*MapW + Cos( 45 )*MapD + Sin( 45 )*MapH )*MINIMAP_SCALE#*Cos( 45 )
	
	If MinimapHUDImage Then FreeImage MinimapHUDImage
	MinimapHUDImage = CreateImage( Width, Height )
	
	MinimapRenderDone = False
	
	MinimapWidth  = Width
	MinimapHeight = Height
	
	MinimapHUDWidth  = 200
	MinimapHUDHeight = 200
	MinimapHUDX = Config\GWidth - MinimapHUDWidth
	MinimapHUDY = Config\GHeight - MinimapHUDHeight
	
	MinimapRenderX = 0
	MinimapRenderY = 0
End Function

Function ProjectVoxelToMinimap( VoxelX#, VoxelY#, VoxelZ# )
	MinimapProjectedX = Floor( MinimapWidth*0.5 + (  Cos( 45 )*VoxelX - Sin( 45 )*VoxelZ                    )*MINIMAP_SCALE# )
	MinimapProjectedY = Floor( MinimapHeight    + ( -Sin( 45 )*VoxelX - Cos( 45 )*VoxelZ - VoxelY )*MINIMAP_SCALE#*Cos( 45 ) - 3.0 )
End Function

Function IncrementalMinimapRender()
	If MinimapRenderDone Then Return
	
	Local Counter = MilliSecs()
	
	SetBuffer ImageBuffer( MinimapHUDImage )
	
	LockBuffer()
	While MilliSecs() - Counter <= 5
		;Local BlockType = RaytraceMinimapPoint( ( MinimapRenderX - MinimapWidth*0.5 )/MINIMAP_SCALE#, ( MinimapRenderY - ( MinimapHeight - Sin( 45 )*MapH*0.5 )*0.5 )/MINIMAP_SCALE# )
		
		RedoMinimapPixel( MinimapRenderX, MinimapRenderY )
		
		MinimapRenderX = MinimapRenderX + 1
		
		If MinimapRenderX = MinimapWidth Then
			MinimapRenderX = 0
			MinimapRenderY = MinimapRenderY + 1
			
			If MinimapRenderY = MinimapHeight Then
				MinimapRenderDone = True
				
				Exit
			EndIf
		EndIf
	Wend
	UnlockBuffer()
	
	SetBuffer BackBuffer()
End Function

Function ChangeMinimapVoxel( VoxelX, VoxelY, VoxelZ )
	If Not MinimapHUDImage Then Return
	
	ProjectVoxelToMinimap( VoxelX, VoxelY, VoxelZ )
	
	Local X = MinimapProjectedX
	Local Y = MinimapProjectedY
	
	For I = 1 To Ceil( MINIMAP_SCALE# )*2
		For S = -I To I
			RedoMinimapPixel( X + S, Y + I )
			RedoMinimapPixel( X + S, Y - I )
			RedoMinimapPixel( X + I, Y + S )
			RedoMinimapPixel( X - I, Y + S )
		Next
	Next
	
	RedoMinimapPixel( X, Y )
End Function

Function RedoMinimapPixel( X, Y )
	Local BlockType = RaytraceMinimapPoint( ( X - MinimapWidth*0.5 )/MINIMAP_SCALE#, ( Y - ( MinimapHeight - Sin( 45 )*MapH*0.5 )*0.5 )/MINIMAP_SCALE# )
	
	Local R = BlockTypes[ BlockType ]\MinimapR
	Local G = BlockTypes[ BlockType ]\MinimapG
	Local B = BlockTypes[ BlockType ]\MinimapB
	
	Local Ratio# = DDA_Y/Float( MapH )*0.8 + 0.2
	
	TFormVector 0, 0, 1, MapLight, 0
	Local Dot# = ( TFormedX()*DDA_NX + TFormedY()*DDA_NY + TFormedZ()*DDA_NZ )
	
	If Dot# < 0.0 Then Dot# = 0.0
	
	Ratio# = Dot#*Ratio#
	
	Ratio# = Ratio# + 0.5
	
	If Ratio# > 1.0 Then Ratio# = 1.0 ElseIf Ratio# < 0.0 Then Ratio# = 0.0
	
	R = R*Ratio#
	G = G*Ratio#
	B = B*Ratio#
	
	If DDA_ThroughWater Then
		R = R*BlockTypes[ BLOCK_WATER ]\MinimapR/256.0
		G = G*BlockTypes[ BLOCK_WATER ]\MinimapG/256.0
		B = B*BlockTypes[ BLOCK_WATER ]\MinimapB/256.0
	EndIf
	
	WritePixel X, Y, ( R Shl 16 ) + ( G Shl 8 ) + B, ImageBuffer( MinimapHUDImage )
End Function

Function RaytraceMinimapPoint( WorldX#, WorldY# )
	Local RayPos#[ 2 ], RayDir#[ 2 ], Grid#[ 2 ], Normal[ 2 ]
	
	WorldY# = -WorldY#
	
	RayPos[ 0 ] = 0.707107*WorldX# + 0.5*WorldY#
	RayPos[ 1 ] = 0.707107*WorldY#
	RayPos[ 2 ] = -0.707107*WorldX# + 0.5*WorldY#
	RayDir[ 0 ] = 0.5
	RayDir[ 1 ] = -0.707107
	RayDir[ 2 ] =  0.5
	
	If RayMapCubeIntersection( RayPos, RayDir, Grid ) Then
		Return DDA( Grid[ 0 ], Grid[ 1 ], Grid[ 2 ], RayDir[ 0 ], RayDir[ 1 ], RayDir[ 2 ] )
	Else
		Return 0
	EndIf
End Function

Function DDA( X#, Y#, Z#, VX#, VY#, VZ# )
	DDA_ThroughWater = False
	
	DirX# = VX#
	DirY# = VY#
	DirZ# = VZ#
	
	InvDirX# = 1.0/DirX#
	InvDirY# = 1.0/DirY#
	InvDirZ# = 1.0/DirZ#

	If DirX# = 0.0 Then InvDirX# = 10000.0
	If DirY# = 0.0 Then InvDirY# = 10000.0
	If DirZ# = 0.0 Then InvDirZ# = 10000.0
	
	GridX# = X#
	GridY# = Y#
	GridZ# = Z#

	VoxelX = Floor( GridX )
	VoxelY = Floor( GridY )
	VoxelZ = Floor( GridZ )

	dTX# = Abs( InvDirX# )
	dTY# = Abs( InvDirY# )
	dTZ# = Abs( InvDirZ# )

	StepX = Sgn( DirX# )
	StepY = Sgn( DirY# )
	StepZ = Sgn( DirZ# )

	If DirX < 0.0 Then MaxTX# = ( GridX - VoxelX )*dTX Else MaxTX# = ( VoxelX + 1.0 - GridX )*dTX
	If DirY < 0.0 Then MaxTY# = ( GridY - VoxelY )*dTY Else MaxTY# = ( VoxelY + 1.0 - GridY )*dTY
	If DirZ < 0.0 Then MaxTZ# = ( GridZ - VoxelZ )*dTZ Else MaxTZ# = ( VoxelZ + 1.0 - GridZ )*dTZ
	
	While VoxelX >= 0 And VoxelY >= 0 And VoxelZ >= 0 And VoxelX < MapW And VoxelY <= MapH And VoxelZ < MapD
		Local BlockType = MapData( VoxelX, VoxelY, VoxelZ )
		
		If BlockType <> 0 And ( Not BlockTypes[ BlockType ]\IsObject ) Then
			If BlockType = BLOCK_WATER Then
				DDA_ThroughWater = True
			Else
				DDA_X = VoxelX
				DDA_Y = VoxelY
				DDA_Z = VoxelZ
				
				Return BlockType
			EndIf
		EndIf

		If MaxTX < MaxTY Then
			If MaxTX < MaxTZ Then
				VoxelX = VoxelX + StepX
				MaxTX  = MaxTX  + dTX
				
				DDA_NX = StepX
				DDA_NY = 0
				DDA_NZ = 0
			Else
				VoxelZ = VoxelZ + StepZ
				MaxTZ  = MaxTZ  + dTZ
				
				DDA_NX = 0
				DDA_NY = 0
				DDA_NZ = StepZ
			EndIf
		Else
			If MaxTY < MaxTZ Then
				VoxelY = VoxelY + StepY
				MaxTY  = MaxTY  + dTY
				
				DDA_NX = 0
				DDA_NY = StepY
				DDA_NZ = 0
			Else
				VoxelZ = VoxelZ + StepZ
				MaxTZ  = MaxTZ  + dTZ
				
				DDA_NX = 0
				DDA_NY = 0
				DDA_NZ = StepZ
			EndIf
		EndIf
	Wend
End Function

Function RayMapCubeIntersection( RayPos#[ 2 ], RayDir#[ 2 ], GridPos#[ 2 ] )
	Local ZeroX# = -MapW*0.5, OneX# = MapW*0.5
	Local ZeroY# = -MapH*0.5, OneY# = MapH*0.5
	Local ZeroZ# = -MapD*0.5, OneZ# = MapD*0.5

	Local PlaneNormal#[ 5 ], PlanePosition#[ 5 ]
	
	PlaneNormal[ 0 ] = -1 : PlaneNormal[ 1 ] = 1
	PlaneNormal[ 2 ] = -1 : PlaneNormal[ 3 ] = 1
	PlaneNormal[ 4 ] = -1 : PlaneNormal[ 5 ] = 1
	
	PlanePosition[ 0 ] = ZeroZ# : PlanePosition[ 1 ] = OneZ#
	PlanePosition[ 2 ] = ZeroY# : PlanePosition[ 3 ] = OneY#
	PlanePosition[ 4 ] = ZeroX# : PlanePosition[ 5 ] = OneX#
	
	Local Indices[ 5 ], Borders#[ 12 ]
	
	Indices[ 0 ] = 0 : Indices[ 1 ] = 1
	Indices[ 2 ] = 0 : Indices[ 3 ] = 2
	Indices[ 4 ] = 1 : Indices[ 5 ] = 2
	
	Borders[ 0  ] = ZeroX# : Borders[ 1  ] = ZeroY#
	Borders[ 2  ] = ZeroX# : Borders[ 3  ] = ZeroZ#
	Borders[ 4  ] = ZeroY# : Borders[ 5  ] = ZeroZ#
	Borders[ 6  ] =  OneX# : Borders[ 7  ] =  OneY#
	Borders[ 8  ] =  OneX# : Borders[ 9  ] =  OneZ#
	Borders[ 10 ] =  OneY# : Borders[ 11 ] =  OneZ#
	
	Local INFINITY# = 10000000.0 ;Not infinity
	Local MinT# = INFINITY#

	For I = 0 To 5
		Local T# = -( PlaneNormal[ I ]*RayPos[ 2 - I/2 ] - PlaneNormal[ I ]*PlanePosition[ I ] )/( PlaneNormal[ I ]*RayDir[ 2 - I/2 ] )
		
		If T# < MinT# Then
			Local Intersection#[ 2 ]
			Intersection[ 0 ] = RayPos[ 0 ] + RayDir[ 0 ]*T#
			Intersection[ 1 ] = RayPos[ 1 ] + RayDir[ 1 ]*T#
			Intersection[ 2 ] = RayPos[ 2 ] + RayDir[ 2 ]*T#
			
			Local A = Intersection[ Indices[ ( ( I Shr 1 ) Shl 1 )     ] ] >= Borders[   ( I Shr 1 ) Shl 1       ]
			Local B = Intersection[ Indices[ ( ( I Shr 1 ) Shl 1 ) + 1 ] ] >= Borders[ ( ( I Shr 1 ) Shl 1 ) + 1 ]
			Local C = Intersection[ Indices[ ( ( I Shr 1 ) Shl 1 )     ] ] <= Borders[ ( ( I Shr 1 ) Shl 1 ) + 6 ]
			Local D = Intersection[ Indices[ ( ( I Shr 1 ) Shl 1 ) + 1 ] ] <= Borders[ ( ( I Shr 1 ) Shl 1 ) + 7 ]

			If A And B And C And D Then
				GridPos[ 0 ] = Intersection[ 0 ] + MapW*0.5
				GridPos[ 1 ] = Intersection[ 1 ] + MapH*0.5
				GridPos[ 2 ] = Intersection[ 2 ] + MapD*0.5
				
				DDA_NX = PlaneNormal[ I ]*( I <= 5 And I >  3 )
				DDA_NY = PlaneNormal[ I ]*( I <= 3 And I >  1 )
				DDA_NZ = PlaneNormal[ I ]*( I <= 1 And I > -1 )

				MinT# = T#
			EndIf
		EndIf
	Next
	
	Return MinT# < INFINITY#
End Function